import ast

from kirin import ir, types, lowering
from kirin.exceptions import DialectLoweringError
from kirin.dialects.py.unpack import unpacking

from .stmts import For, Yield, IfElse
from ._dialect import dialect


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_If(self, state: lowering.LoweringState, node: ast.If) -> lowering.Result:
        cond = state.visit(node.test).expect_one()
        frame = state.current_frame
        body_frame = lowering.Frame.from_stmts(node.body, state, globals=frame.globals)
        then_cond = body_frame.curr_block.args.append_from(types.Bool, cond.name)
        if cond.name:
            body_frame.defs[cond.name] = then_cond
        state.push_frame(body_frame)
        state.exhaust(body_frame)
        state.pop_frame(finalize_next=False)  # NOTE: scf does not have multiple blocks

        else_frame = lowering.Frame.from_stmts(
            node.orelse, state, globals=frame.globals
        )
        else_cond = else_frame.curr_block.args.append_from(types.Bool, cond.name)
        if cond.name:
            else_frame.defs[cond.name] = else_cond
        state.push_frame(else_frame)
        state.exhaust(else_frame)
        state.pop_frame(finalize_next=False)  # NOTE: scf does not have multiple blocks

        yield_names: list[str] = []
        body_yields: list[ir.SSAValue] = []
        else_yields: list[ir.SSAValue] = []
        if node.orelse:
            for name in body_frame.defs.keys():
                if name in else_frame.defs:
                    yield_names.append(name)
                    body_yields.append(body_frame.get_scope(name))
                    else_yields.append(else_frame.get_scope(name))
        else:
            for name in body_frame.defs.keys():
                if name in frame.defs:
                    yield_names.append(name)
                    body_yields.append(body_frame.get_scope(name))
                    value = frame.get(name)
                    if value is None:
                        raise DialectLoweringError(f"expected value for {name}")
                    else_yields.append(value)

        if not (
            body_frame.curr_block.last_stmt
            and body_frame.curr_block.last_stmt.has_trait(ir.IsTerminator)
        ):
            body_frame.append_stmt(Yield(*body_yields))

        if not (
            else_frame.curr_block.last_stmt
            and else_frame.curr_block.last_stmt.has_trait(ir.IsTerminator)
        ):
            else_frame.append_stmt(Yield(*else_yields))

        stmt = IfElse(
            cond,
            then_body=body_frame.curr_region,
            else_body=else_frame.curr_region,
        )
        for result, name, body, else_ in zip(
            stmt.results, yield_names, body_yields, else_yields
        ):
            result.name = name
            result.type = body.type.join(else_.type)
            frame.defs[name] = result
        state.append_stmt(stmt)
        return lowering.Result()

    def lower_For(
        self, state: lowering.LoweringState, node: ast.For
    ) -> lowering.Result:
        iter_ = state.visit(node.iter).expect_one()

        yields: list[str] = []

        def new_block_arg_if_inside_loop(frame: lowering.Frame, capture: ir.SSAValue):
            if not capture.name:
                raise DialectLoweringError("unexpected loop variable captured")
            yields.append(capture.name)
            return frame.curr_block.args.append_from(capture.type, capture.name)

        body_frame = state.push_frame(
            lowering.Frame.from_stmts(
                node.body,
                state,
                globals=state.current_frame.globals,
                capture_callback=new_block_arg_if_inside_loop,
            )
        )
        loop_var = body_frame.curr_block.args.append_from(types.Any)
        unpacking(state, node.target, loop_var)
        state.exhaust(body_frame)
        # NOTE: this frame won't have phi nodes
        if yields and (
            body_frame.curr_block.last_stmt is None
            or not body_frame.curr_block.last_stmt.has_trait(ir.IsTerminator)
        ):
            body_frame.append_stmt(Yield(*[body_frame.defs[name] for name in yields]))  # type: ignore
        state.pop_frame(finalize_next=False)  # NOTE: scf does not have multiple blocks

        initializers: list[ir.SSAValue] = []
        for name in yields:
            value = state.current_frame.get(name)
            if value is None:
                raise DialectLoweringError(f"expected value for {name}")
            initializers.append(value)
        stmt = For(iter_, body_frame.curr_region, *initializers)
        for name, result in zip(yields, stmt.results):
            state.current_frame.defs[name] = result
            result.name = name
        state.append_stmt(stmt)
        return lowering.Result()



import re
import sys
import inspect
import dataclasses
from types import ModuleType
from typing import Callable, get_args
from dataclasses import KW_ONLY

from beartype.door import is_subhint

from kirin import ir, types

from .base import BaseModifier
from .info import Field, ArgumentField, AttributeField, StatementFields, argument


class ScanFields(BaseModifier):
    _FIELDS = "__kirin_stmt_fields"
    # String regex that string annotations for ClassVar or InitVar must match.
    # Allows "identifier.identifier[" or "identifier[".
    # https://bugs.python.org/issue33453 for details.
    _MODULE_IDENTIFIER_RE = re.compile(r"^(?:\s*(\w+)\s*\.)?\s*(\w+)")

    def scan_fields(self):
        self._scan_base_fields()
        self._scan_annotations()
        setattr(self.cls, self._FIELDS, self.fields)
        return

    def _scan_base_fields(self):
        # let's just assume dict preserve insertion order in python 3.7+
        for b in self.cls.__mro__[-1:0:-1]:  # taken from dataclass impl
            base_fields: StatementFields | None = getattr(b, self._FIELDS, None)
            if base_fields is not None:
                self.has_statement_bases = True
                for f in base_fields:
                    assert f.name is not None, "field name must be set"
                    self.fields[f.name] = f

    def _scan_annotations(self):
        cls_fields: list[Field] = []
        cls_annotations = inspect.get_annotations(self.cls, eval_str=True)
        for name, typ in cls_annotations.items():
            # See if this is a marker to change the value of kw_only.
            if self._is_kw_only(typ) or (
                isinstance(typ, str)
                and self._is_type(typ, dataclasses, KW_ONLY, self._is_kw_only)
            ):
                if self.KW_ONLY_seen:
                    raise TypeError(
                        f"{name!r} is KW_ONLY, but KW_ONLY "
                        "has already been specified"
                    )
                self.KW_ONLY_seen = True
                self.kw_only = True
            else:
                cls_fields.append(self._get_field(name, typ))

        for f in cls_fields:
            name: str = f.name  # type: ignore # name has been set
            self.fields[name] = f
            if hasattr(self.cls, name):
                # remove the field from the class
                # unlike dataclass, we don't actually
                # store any values in the class, they
                # are stored inside the IR node.
                delattr(self.cls, name)

    def _get_field(self, name: str, typ):
        """Return a Field object for this field name and type."""
        guess = self._guess_type(typ)
        default = getattr(self.cls, name, dataclasses.MISSING)
        if isinstance(default, Field):
            f = default
        # no default descriptor, create a new one
        elif is_subhint(typ, ir.SSAValue) or is_subhint(typ, tuple[ir.SSAValue, ...]):
            f: Field = (
                argument()
            )  # only argument can have default, others must use field specifiers
        elif typ is ir.ResultValue:
            raise ValueError(
                f"expect field specifiers for `ir.ResultValue` fields: {name}"
            )
        elif typ is ir.Region:
            raise ValueError(f"expect field specifiers for `ir.Region` fields: {name}")
        elif typ is ir.Block:
            raise ValueError(f"expect field specifiers for `ir.Block` fields: {name}")
        else:
            # could be a wrong SSAValue field, e.g list[SSAValue], try check args
            if typ_args := get_args(typ):
                if any(is_subhint(arg, ir.SSAValue) for arg in typ_args):
                    raise ValueError(
                        f"unsupported SSAValue field: {name},"
                        f" expect `SSAValue` or `tuple[SSAValue, ...]`, got {typ}"
                    )
            raise ValueError(f"expect field specifiers for attribute fields: {name}")

        f.name = name
        f.annotation = typ
        self._post_init_field(f, guess)
        return f

    def _post_init_field(self, f: Field, guess: type | None):
        if isinstance(f, ArgumentField) and is_subhint(guess, tuple[ir.SSAValue, ...]):
            f.group = True
        # try to narrow the type based on the guess
        elif isinstance(f, AttributeField):
            if guess and not is_subhint(
                guess, ir.Attribute
            ):  # not specified, and using python type
                if f.type is types.Any:  # not set or too generic
                    f.type = types.hint2type(guess)
                f.pytype = True

    @staticmethod
    def _is_kw_only(a_type):
        return a_type is KW_ONLY

    def _is_type(
        self,
        annotation: type | str,
        obj_module: ModuleType,
        obj_type: type,
        is_type_predicate: Callable[[type], bool],
    ):
        """Given a type annotation string, does it refer to `obj_type` in
        `obj_module`?  For example, when checking that annotation denotes a
        `ClassVar`, then `obj_module` is typing, and a_type is
        `typing.ClassVar`.

        Taken from dataclasses.py/is_type.
        """
        guess = self._guess_type(annotation)
        if guess is not None:
            return guess.__module__ is obj_module and is_type_predicate(guess)
        return False

    def _guess_type(self, annotation: type | str) -> type | None:
        """Guess the type/hint object from a string annotation."""
        if not isinstance(annotation, str):
            return annotation

        match = self._MODULE_IDENTIFIER_RE.match(annotation)
        ns = None
        if match:
            module_name = match.group(1)
            if not module_name:
                # No module name, assume the class's module did
                # "from dataclasses import InitVar".
                ns = sys.modules.get(self.cls.__module__).__dict__
            else:
                # Look up module_name in the class's module.
                if self.cls_module:
                    ns = self.cls_module.__dict__.get(module_name).__dict__
            if ns:
                return ns.get(match.group(2))
        return None



from collections.abc import Iterable

from kirin import ir, interp
from kirin.analysis import const
from kirin.dialects import func

from .stmts import For, Yield, IfElse
from ._dialect import dialect

# NOTE: unlike concrete interpreter, we need to use a new frame
# for each iteration because otherwise join two constant values
# will result in bottom (error) element.


@dialect.register(key="constprop")
class DialectConstProp(interp.MethodTable):

    @interp.impl(Yield)
    def yield_stmt(
        self,
        interp_: const.Propagate,
        frame: const.Frame,
        stmt: Yield,
    ):
        return interp.YieldValue(frame.get_values(stmt.values))

    @interp.impl(IfElse)
    def if_else(
        self,
        interp_: const.Propagate,
        frame: const.Frame,
        stmt: IfElse,
    ):
        cond = frame.get(stmt.cond)
        if isinstance(cond, const.Value):
            if cond.data:
                body = stmt.then_body
            else:
                body = stmt.else_body
            body_frame, ret = self._prop_const_cond_ifelse(
                interp_, frame, stmt, cond, body
            )
            frame.entries.update(body_frame.entries)
            if not body_frame.frame_is_not_pure and not isinstance(
                body.blocks[0].last_stmt, func.Return
            ):
                frame.should_be_pure.add(stmt)
            return ret
        else:
            then_frame, then_results = self._prop_const_cond_ifelse(
                interp_, frame, stmt, const.Value(True), stmt.then_body
            )
            else_frame, else_results = self._prop_const_cond_ifelse(
                interp_, frame, stmt, const.Value(False), stmt.else_body
            )
            # NOTE: then_frame and else_frame do not change
            # parent frame variables value except cond
            frame.entries.update(then_frame.entries)
            frame.entries.update(else_frame.entries)
            # TODO: pick the non-return value
            if isinstance(then_results, interp.ReturnValue) and isinstance(
                else_results, interp.ReturnValue
            ):
                return interp.ReturnValue(then_results.value.join(else_results.value))
            elif isinstance(then_results, interp.ReturnValue):
                ret = else_results
            elif isinstance(else_results, interp.ReturnValue):
                ret = then_results
            else:
                if not (
                    then_frame.frame_is_not_pure is True
                    or else_frame.frame_is_not_pure is True
                ):
                    frame.should_be_pure.add(stmt)
                ret = interp_.join_results(then_results, else_results)
        return ret

    def _prop_const_cond_ifelse(
        self,
        interp_: const.Propagate,
        frame: const.Frame,
        stmt: IfElse,
        cond: const.Value,
        body: ir.Region,
    ):
        with interp_.state.new_frame(interp_.new_frame(stmt)) as body_frame:
            body_frame.entries.update(frame.entries)
            body_frame.set(body.blocks[0].args[0], cond)
            results = interp_.run_ssacfg_region(body_frame, body)
        return body_frame, results

    @interp.impl(For)
    def for_loop(
        self,
        interp_: const.Propagate,
        frame: const.Frame,
        stmt: For,
    ):
        iterable = frame.get(stmt.iterable)
        if isinstance(iterable, const.Value):
            return self._prop_const_iterable_forloop(interp_, frame, stmt, iterable)
        else:  # TODO: support other iteration
            return tuple(interp_.lattice.top() for _ in stmt.results)

    def _prop_const_iterable_forloop(
        self,
        interp_: const.Propagate,
        frame: const.Frame,
        stmt: For,
        iterable: const.Value,
    ):
        frame_is_not_pure = False
        if not isinstance(iterable.data, Iterable):
            raise interp.InterpreterError(
                f"Expected iterable, got {type(iterable.data)}"
            )

        loop_vars = frame.get_values(stmt.initializers)
        body_block = stmt.body.blocks[0]
        block_args = body_block.args

        for value in iterable.data:
            with interp_.state.new_frame(interp_.new_frame(stmt)) as body_frame:
                body_frame.entries.update(frame.entries)
                body_frame.set_values(
                    block_args,
                    (const.Value(value),) + loop_vars,
                )
                loop_vars = interp_.run_ssacfg_region(body_frame, stmt.body)

            if body_frame.frame_is_not_pure:
                frame_is_not_pure = True
            if loop_vars is None:
                loop_vars = ()
            elif isinstance(loop_vars, interp.ReturnValue):
                return loop_vars

        if not frame_is_not_pure:
            frame.should_be_pure.add(stmt)
        return loop_vars



from typing import TYPE_CHECKING, Generic, TypeVar, Callable, ParamSpec
from dataclasses import dataclass

if TYPE_CHECKING:
    from kirin.ir.nodes.stmt import Statement

Params = ParamSpec("Params")
RetType = TypeVar("RetType")


@dataclass(frozen=True)
class Binding(Generic[Params, RetType]):
    parent: type["Statement"]

    def __call__(self, *args: Params.args, **kwargs: Params.kwargs) -> RetType:
        raise NotImplementedError(
            f"Binding of {self.parent.name} can \
            only be called from a kernel"
        )


def wraps(parent: type["Statement"]):
    """Wraps a [`Statement`][kirin.ir.nodes.stmt.Statement] to a `Binding` object
    which will be special cased in the lowering process.

    This is useful for providing type hints by faking the call signature of a
    [`Statement`][kirin.ir.nodes.stmt.Statement].

    ## Example

    Directly writing a function with the statement will let Python linter think
    you intend to call the constructor of the statement class. However, given the
    context of a kernel, our intention is to actually "call" the statement, e.g
    the following will produce type errors with pyright or mypy:

    ```python
    from kirin.dialects import math
    from kirin.prelude import basic_no_opt

    @basic_no_opt
    def main(x: float):
        return math.sin(x) # this is a statement, not a function
    ```

    the `@lowering.wraps` decorator allows us to provide a type hint for the
    statement, e.g:

    ```python
    from kirin import lowering

    @lowering.wraps(math.sin)
    def sin(value: float) -> float: ...

    @basic_no_opt
    def main(x: float):
        return sin(x) # linter now thinks this is a function

    sin(1.0) # this will raise a NotImplementedError("Binding of sin can only be called from a kernel")
    ```
    """

    def wrapper(func: Callable[Params, RetType]) -> Binding[Params, RetType]:
        return Binding(parent)

    return wrapper



from kirin import types
from kirin.interp import Frame, MethodTable, impl
from kirin.dialects.eltype import ElType
from kirin.dialects.py.binop import Add
from kirin.analysis.typeinfer import TypeInference
from kirin.dialects.py.indexing import GetItem

from .stmts import New, Push, Range, IListType
from .runtime import IList
from ._dialect import dialect


@dialect.register(key="typeinfer")
class TypeInfer(MethodTable):

    @impl(Range)
    def range(
        self, interp_: TypeInference, frame: Frame[types.TypeAttribute], stmt: Range
    ):
        result = interp_.maybe_const(stmt.result, IList)
        if result:
            return (IListType[types.Int, types.Literal(len(result))],)
        return (IListType[types.Int, types.Any],)

    @staticmethod
    def _get_list_len(typ: types.Generic):
        if isinstance(typ.vars[1], types.Literal) and isinstance(typ.vars[1].data, int):
            return typ.vars[1].data
        else:
            return types.Any

    @impl(ElType, types.PyClass(IList))
    def eltype_list(
        self, interp: TypeInference, frame: Frame[types.TypeAttribute], stmt: ElType
    ):
        list_type = frame.get(stmt.container)
        if isinstance(list_type, types.Generic):
            return (list_type.vars[0],)
        else:
            return (types.Any,)

    @impl(New)
    def new(self, interp: TypeInference, frame: Frame[types.TypeAttribute], stmt: New):
        values = frame.get_values(stmt.values)
        if not values:
            return (IListType[types.Any, types.Literal(0)],)

        elem_type = values[0]
        for v in values:
            elem_type = elem_type.join(v)

        return (IListType[elem_type, types.Literal(len(values))],)

    @impl(Push)
    def push(
        self, interp: TypeInference, frame: Frame[types.TypeAttribute], stmt: Push
    ):
        lst_type: types.Generic = frame.get(stmt.lst)  # type: ignore
        value_type = frame.get(stmt.value)
        if not lst_type.is_subseteq(IListType):
            return (types.Bottom,)

        if not lst_type.vars[0].is_subseteq(value_type):
            return (types.Bottom,)

        lst_len = self._get_list_len(lst_type)
        if not isinstance(lst_len, int):
            return (IListType[lst_type.vars[0], types.Any],)

        return (IListType[lst_type.vars[0], types.Literal(lst_len + 1)],)

    @impl(Add, types.PyClass(IList), types.PyClass(IList))
    def add(self, interp: TypeInference, frame: Frame[types.TypeAttribute], stmt: Add):
        lhs_type = frame.get(stmt.lhs)
        rhs_type = frame.get(stmt.rhs)
        if not lhs_type.is_subseteq(IListType) or not rhs_type.is_subseteq(IListType):
            return (types.Bottom,)

        if not isinstance(lhs_type, types.Generic):  # just annotated with list
            lhs_type = IListType[types.Any, types.Any]

        if not isinstance(rhs_type, types.Generic):
            rhs_type = IListType[types.Any, types.Any]

        if len(lhs_type.vars) != 2 or len(rhs_type.vars) != 2:
            raise TypeError("missing type argument for list")

        elem_type = lhs_type.vars[0].join(rhs_type.vars[0])

        lhs_len = self._get_list_len(lhs_type)
        rhs_len = self._get_list_len(rhs_type)
        if isinstance(lhs_len, int) and isinstance(rhs_len, int):
            return (IListType[elem_type, types.Literal(lhs_len + rhs_len)],)
        return (IListType[elem_type, types.Any],)

    @impl(GetItem, types.PyClass(IList), types.PyClass(int))
    def getitem(
        self, interp: TypeInference, frame: Frame[types.TypeAttribute], stmt: GetItem
    ):
        obj_type = frame.get(stmt.obj)
        if not obj_type.is_subseteq(IListType):
            raise TypeError(f"Expected list, got {obj_type}")

        # just list type
        if not isinstance(obj_type, types.Generic):
            return (types.Any,)
        else:
            return (obj_type.vars[0],)

    @impl(GetItem, types.PyClass(IList), types.PyClass(slice))
    def getitem_slice(
        self, interp: TypeInference, frame: Frame[types.TypeAttribute], stmt: GetItem
    ):
        obj_type = frame.get(stmt.obj)
        if not obj_type.is_subseteq(IListType):
            raise TypeError(f"Expected list, got {obj_type}")

        # just list type
        if not isinstance(obj_type, types.Generic):
            return (IListType[types.Any, types.Any],)
        elif index_ := interp.maybe_const(stmt.index, slice):
            # TODO: actually calculate the size
            obj_len = obj_type.vars[1]
            if not isinstance(obj_len, types.Literal):
                return (IListType[obj_type.vars[0], types.Any],)
            LenT = types.Literal(len(range(obj_len.data)[index_]))
            return (IListType[obj_type.vars[0], LenT],)
        else:
            return (IListType[obj_type.vars[0], types.Any],)



from dataclasses import dataclass

from kirin import ir
from kirin.analysis import const
from kirin.rewrite.abc import RewriteRule, RewriteResult


@dataclass
class WrapConst(RewriteRule):
    """Insert constant hints into the SSA values.

    !!! note
        This pass is not exactly the same as ConstantFold. ConstantFold
        only rewrites the SSAValue with `const.Value` into a constant
        statement. This rule, however, inserts the entire constant lattice
        into the SSAValue's hints. Thus enabling other rules/analysis to
        utilize the constant information beyond just `const.Value`.
    """

    frame: const.Frame

    def wrap(self, value: ir.SSAValue) -> bool:
        result = self.frame.entries.get(value)
        if not result:
            return False

        const_hint = value.hints.get("const")
        if const_hint and isinstance(const_hint, const.Result):
            const_result = result.meet(const_hint)
            if const_result.is_equal(const_hint):
                return False
        else:
            const_result = result
        value.hints["const"] = const_result
        return True

    def rewrite_Block(self, node: ir.Block) -> RewriteResult:
        has_done_something = False
        for arg in node.args:
            if self.wrap(arg):
                has_done_something = True
        return RewriteResult(has_done_something=has_done_something)

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        has_done_something = False
        for result in node.results:
            if self.wrap(result):
                has_done_something = True

        if (
            trait := node.get_trait(ir.MaybePure)
        ) and node in self.frame.should_be_pure:
            trait.set_pure(node)
            has_done_something = True
        return RewriteResult(has_done_something=has_done_something)



from __future__ import annotations

from typing import Iterable, Sequence, overload
from dataclasses import field, dataclass

from kirin.ir import SSAValue, Statement
from kirin.exceptions import DialectLoweringError


@dataclass(init=False)
class Result(Sequence[SSAValue]):
    values: Sequence[SSAValue] = field(default_factory=list)

    @overload
    def __init__(self, value: None = None) -> None: ...

    @overload
    def __init__(self, value: SSAValue, *values: SSAValue) -> None: ...

    @overload
    def __init__(self, value: Iterable[SSAValue]) -> None: ...

    @overload
    def __init__(self, value: Statement) -> None: ...

    def __init__(
        self,
        value: SSAValue | Iterable[SSAValue] | Statement | None = None,
        *values: SSAValue,
    ) -> None:
        if value is None:
            assert not values, "unexpected values"
            self.values = []
        elif isinstance(value, SSAValue):
            self.values = [value, *values]
        elif isinstance(value, Statement):
            assert not values, "unexpected values"
            self.values = value._results
        else:
            assert not values, "unexpected values"
            self.values = list(value)

    def expect_one(self) -> SSAValue:
        if len(self.values) != 1:
            raise DialectLoweringError("expected one result")
        return self.values[0]

    # forward the sequence methods
    def __len__(self):
        return len(self.values)

    def __getitem__(self, key):
        return self.values[key]

    def __iter__(self):
        return iter(self.values)

    def __contains__(self, value):
        return value in self.values

    def __reversed__(self):
        return reversed(self.values)

    def __eq__(self, other):
        return self.values == other.values

    def __ne__(self, other):
        return self.values != other.values



from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar, Iterable
from dataclasses import field, dataclass

from typing_extensions import Self

from kirin.ir import SSAValue, Statement

from .exceptions import InterpreterError

ValueType = TypeVar("ValueType")


@dataclass
class FrameABC(ABC, Generic[ValueType]):
    """Abstract base class for interpreter frame."""

    code: Statement
    """func statement being interpreted.
    """

    @classmethod
    @abstractmethod
    def from_func_like(cls, code: Statement) -> Self:
        """Create a new frame for the given method."""
        ...

    @abstractmethod
    def get(self, key: SSAValue) -> ValueType:
        """Get the value for the given [`SSAValue`][kirin.ir.SSAValue] key.
        See also [`get_values`][kirin.interp.frame.Frame.get_values].

        Args:
            key(SSAValue): The key to get the value for.

        Returns:
            ValueType: The value.
        """
        ...

    @abstractmethod
    def set(self, key: SSAValue, value: ValueType) -> None:
        """Set the value for the given [`SSAValue`][kirin.ir.SSAValue] key.
        See also [`set_values`][kirin.interp.frame.Frame.set_values].

        Args:
            key(SSAValue): The key to set the value for.
            value(ValueType): The value.
        """
        ...

    def get_values(self, keys: Iterable[SSAValue]) -> tuple[ValueType, ...]:
        """Get the values of the given [`SSAValue`][kirin.ir.SSAValue] keys.
        See also [`get`][kirin.interp.frame.Frame.get].

        Args:
            keys(Iterable[SSAValue]): The keys to get the values for.

        Returns:
            tuple[ValueType, ...]: The values.
        """
        return tuple(self.get(key) for key in keys)

    def set_values(self, keys: Iterable[SSAValue], values: Iterable[ValueType]) -> None:
        """Set the values of the given [`SSAValue`][kirin.ir.SSAValue] keys.
        This is a convenience method to set multiple values at once.

        Args:
            keys(Iterable[SSAValue]): The keys to set the values for.
            values(Iterable[ValueType]): The values.
        """
        for key, value in zip(keys, values):
            self.set(key, value)

    @abstractmethod
    def set_stmt(self, stmt: Statement) -> Self:
        """Set the current statement."""
        ...


@dataclass
class Frame(FrameABC[ValueType]):
    """Interpreter frame."""

    lino: int = 0
    stmt: Statement | None = None
    """statement being interpreted.
    """

    globals: dict[str, Any] = field(default_factory=dict)
    """Global variables this frame has access to.
    """

    # NOTE: we are sharing the same frame within blocks
    # this is because we are validating e.g SSA value pointing
    # to other blocks separately. This avoids the need
    # to have a separate frame for each block.
    entries: dict[SSAValue, ValueType] = field(default_factory=dict)
    """SSA values and their corresponding values.
    """

    @classmethod
    def from_func_like(cls, code: Statement) -> Self:
        """Create a new frame for the given statement."""
        return cls(code=code)

    def get(self, key: SSAValue) -> ValueType:
        """Get the value for the given [`SSAValue`][kirin.ir.SSAValue].

        Args:
            key(SSAValue): The key to get the value for.

        Returns:
            ValueType: The value.

        Raises:
            InterpreterError: If the value is not found. This will be catched by the interpreter.
        """
        err = InterpreterError(f"SSAValue {key} not found")
        value = self.entries.get(key, err)
        if isinstance(value, InterpreterError):
            raise err
        else:
            return value

    ExpectedType = TypeVar("ExpectedType")

    def get_typed(self, key: SSAValue, type_: type[ExpectedType]) -> ExpectedType:
        """Similar to [`get`][kirin.interp.frame.Frame.get] but also checks the type.

        Args:
            key(SSAValue): The key to get the value for.
            type_(type): The expected type.

        Returns:
            ExpectedType: The value.

        Raises:
            InterpreterError: If the value is not of the expected type.
        """
        value = self.get(key)
        if not isinstance(value, type_):
            raise InterpreterError(f"expected {type_}, got {type(value)}")
        return value

    def set(self, key: SSAValue, value: ValueType) -> None:
        self.entries[key] = value

    def set_stmt(self, stmt: Statement) -> Self:
        self.stmt = stmt
        return self



"""A minimal language example with a single pass that does nothing."""

from kirin import ir
from kirin.dialects import cf, py, func, lowering


@ir.dialect_group(
    [
        func,
        lowering.func,
        lowering.call,
        lowering.cf,
        py.base,
        py.constant,
        py.assign,
        py.binop,
        py.unary,
    ]
)
def simple(self):
    def run_pass(mt):
        return mt

    return run_pass


@simple
def main(x):
    y = x + 1
    return y


main.print()


@simple.add(cf).add(py.cmp)
def main2(x):
    y = x + 1
    if y > 0:  # errors
        return y
    else:
        return -y


main2.print()



from abc import ABC, abstractmethod
from typing import Generic, TypeVar, Iterator, Sequence, overload
from dataclasses import dataclass

from typing_extensions import Self

from kirin.ir.ssa import SSAValue
from kirin.ir.nodes.base import IRNode

ElemType = TypeVar("ElemType", bound=IRNode | SSAValue)
FieldType = TypeVar("FieldType", bound=Sequence)
NodeType = TypeVar("NodeType", bound=IRNode | SSAValue)


@dataclass
class View(ABC, Generic[NodeType, ElemType]):
    node: NodeType

    @abstractmethod
    def __iter__(self) -> Iterator[ElemType]: ...

    @abstractmethod
    def __len__(self) -> int: ...

    def __bool__(self) -> bool:
        return bool(len(self))

    def append(self, value: ElemType) -> None:
        raise NotImplementedError

    def extend(self, values: Sequence[ElemType]) -> None:
        for value in values:
            self.append(value)

    def __reversed__(self) -> Iterator[ElemType]:
        raise NotImplementedError


@dataclass
class SequenceView(Generic[FieldType, NodeType, ElemType], View[NodeType, ElemType]):
    field: FieldType

    @classmethod
    def similar(cls, node: NodeType, field: FieldType) -> Self:
        return cls(node, field)

    def __iter__(self) -> Iterator[ElemType]:
        return iter(self.field)

    def __len__(self) -> int:
        return len(self.field)

    def __reversed__(self) -> Iterator[ElemType]:
        return reversed(self.field)

    def isempty(self) -> bool:
        return len(self) == 0

    def __bool__(self) -> bool:
        return not self.isempty()

    # optional interface
    @overload
    def __getitem__(self, idx: int) -> ElemType: ...

    @overload
    def __getitem__(self, idx: slice) -> Self: ...

    def __getitem__(self, idx: int | slice) -> ElemType | Self:
        if isinstance(idx, slice):
            x: FieldType = self.field[idx]  # type: ignore
            return self.similar(self.node, x)
        else:
            return self.field[idx]


@dataclass
class MutableSequenceView(SequenceView[FieldType, NodeType, ElemType]):
    @overload
    def __setitem__(self, idx: int, value: ElemType) -> None: ...

    @overload
    def __setitem__(self, idx: slice, value: Sequence[ElemType]) -> None: ...

    def __setitem__(
        self, idx: int | slice, value: ElemType | Sequence[ElemType]
    ) -> None:
        if isinstance(idx, int) and not isinstance(value, Sequence):
            return self.set_item(idx, value)
        elif isinstance(idx, slice):
            assert isinstance(value, Sequence), "Expected sequence of values"
            if idx.step is not None:  # no need to support step
                raise ValueError("Slice step is not supported")
            return self.set_item_slice(idx, value)
        else:
            raise TypeError("Expected int or slice")

    def set_item(self, idx: int, value: ElemType) -> None:
        raise NotImplementedError

    def set_item_slice(self, s: slice, value: Sequence[ElemType]) -> None:
        # replace the view of slice
        for idx in range(s.start, s.stop):
            if idx < len(value):
                self.set_item(idx, value[idx])
            else:
                del self[idx]

        # insert the rest of the values
        for idx, v in enumerate(value[s.stop - s.start :]):
            self.insert(idx + s.stop, v)

    def __delitem__(self, idx: int) -> None:
        raise NotImplementedError

    def insert(self, idx: int, value: ElemType) -> None:
        raise NotImplementedError

    def pop(self, idx: int = -1) -> ElemType:
        item = self.field[idx]
        del self[idx]
        return item

    def poplast(self) -> ElemType | None:
        """Pop the last element from the view.

        Returns:
            The last element in the view.
        """
        if self:
            return self.pop(-1)
        return None

    def popfirst(self) -> ElemType | None:
        """Pop the first element from the view.

        Returns:
            The first element in the view.
        """
        if self:
            return self.pop(0)
        return None



from typing import TYPE_CHECKING
from dataclasses import dataclass

from kirin.exceptions import VerificationError
from kirin.ir.attrs.py import PyAttr
from kirin.ir.traits.abc import StmtTrait

if TYPE_CHECKING:
    from kirin.ir import Statement


@dataclass(frozen=True)
class SymbolOpInterface(StmtTrait):
    """A trait that indicates that a statement is a symbol operation.

    A symbol operation is a statement that has a symbol name attribute.
    """

    def get_sym_name(self, stmt: "Statement") -> "PyAttr[str]":
        sym_name: PyAttr[str] | None = stmt.get_attr_or_prop("sym_name")  # type: ignore
        # NOTE: unlike MLIR or xDSL we do not allow empty symbol names
        if sym_name is None:
            raise ValueError(f"Statement {stmt.name} does not have a symbol name")
        return sym_name

    def verify(self, stmt: "Statement"):
        from kirin.types import String

        sym_name = self.get_sym_name(stmt)
        if not (isinstance(sym_name, PyAttr) and sym_name.type.is_subseteq(String)):
            raise ValueError(f"Symbol name {sym_name} is not a string attribute")


@dataclass(frozen=True)
class SymbolTable(StmtTrait):
    """
    Statement with SymbolTable trait can only have one region with one block.
    """

    @staticmethod
    def walk(stmt: "Statement"):
        return stmt.regions[0].blocks[0].stmts

    def verify(self, stmt: "Statement"):
        if len(stmt.regions) != 1:
            raise VerificationError(
                stmt,
                f"Statement {stmt.name} with SymbolTable trait must have exactly one region",
            )

        if len(stmt.regions[0].blocks) != 1:
            raise VerificationError(
                stmt,
                f"Statement {stmt.name} with SymbolTable trait must have exactly one block",
            )

        # TODO: check uniqueness of symbol names



from typing import IO, TypeVar

from kirin import emit
from kirin.interp import MethodTable, InterpreterError, impl
from kirin.emit.julia import EmitJulia

from .stmts import Call, Invoke, Lambda, Return, Function, GetField, ConstantNone
from .dialect import dialect

IO_t = TypeVar("IO_t", bound=IO)


@dialect.register(key="emit.julia")
class JuliaMethodTable(MethodTable):

    @impl(Function)
    def emit_function(
        self, interp: EmitJulia[IO_t], frame: emit.EmitStrFrame, stmt: Function
    ):
        fn_args = stmt.body.blocks[0].args[1:]
        argnames = frame.get_values(fn_args)
        argtypes = tuple(interp.emit_attribute(x.type) for x in fn_args)
        args = [f"{name}::{type}" for name, type in zip(argnames, argtypes)]
        interp.write(f"function {stmt.sym_name}({', '.join(args)})")
        frame.indent += 1
        interp.run_ssacfg_region(frame, stmt.body)
        frame.indent -= 1
        interp.writeln(frame, "end")
        return ()

    @impl(Return)
    def emit_return(
        self, interp: EmitJulia[IO_t], frame: emit.EmitStrFrame, stmt: Return
    ):
        interp.writeln(frame, f"return {frame.get(stmt.value)}")
        return ()

    @impl(ConstantNone)
    def emit_constant_none(
        self, interp: EmitJulia[IO_t], frame: emit.EmitStrFrame, stmt: ConstantNone
    ):
        return ("nothing",)

    @impl(Call)
    def emit_call(self, interp: EmitJulia[IO_t], frame: emit.EmitStrFrame, stmt: Call):
        if stmt.kwargs:
            raise InterpreterError("cannot emit kwargs for dyanmic calls")
        return (
            f"{frame.get(stmt.callee)}({', '.join(frame.get_values(stmt.inputs))})",
        )

    @impl(Invoke)
    def emit_invoke(
        self, interp: EmitJulia[IO_t], frame: emit.EmitStrFrame, stmt: Invoke
    ):
        args = interp.permute_values(
            stmt.callee.arg_names, frame.get_values(stmt.inputs), stmt.kwargs
        )
        return (f"{stmt.callee.sym_name}({', '.join(args)})",)

    @impl(Lambda)
    def emit_lambda(
        self, interp: EmitJulia[IO_t], frame: emit.EmitStrFrame, stmt: Lambda
    ):
        args = tuple(interp.ssa_id[x] for x in stmt.body.blocks[0].args[1:])
        frame.set_values(stmt.body.blocks[0].args, args)
        frame.set_values((stmt.body.blocks[0].args[0],), (stmt.sym_name,))
        frame.captured[stmt.body.blocks[0].args[0]] = frame.get_values(stmt.captured)
        interp.writeln(frame, f"function {stmt.sym_name}({', '.join(args[1:])})")
        frame.indent += 1
        interp.run_ssacfg_region(frame, stmt.body)
        frame.indent -= 1
        interp.writeln(frame, "end")
        return (stmt.sym_name,)

    @impl(GetField)
    def emit_getfield(
        self, interp: EmitJulia[IO_t], frame: emit.EmitStrFrame, stmt: GetField
    ):
        return (frame.captured[stmt.obj][stmt.field],)



from io import StringIO
from typing import cast
from dataclasses import field, dataclass

import stmts
from group import food
from dialect import dialect

from kirin import ir, interp
from lattice import Item, ItemServing, AtLeastXItem, ConstIntItem
from kirin.emit import EmitStr, EmitStrFrame
from kirin.dialects import func
from kirin.emit.exceptions import EmitError


def default_menu_price():
    return {
        "burger": 3.0,
        "salad": 4.0,
        "chicken": 2.0,
    }


@dataclass
class EmitReceptMain(EmitStr):
    keys = ["emit.recept"]
    dialects: ir.DialectGroup = field(default_factory=lambda: food)
    file: StringIO = field(default_factory=StringIO)
    menu_price: dict[str, float] = field(default_factory=default_menu_price)
    recept_analysis_result: dict[ir.SSAValue, Item] = field(default_factory=dict)

    def initialize(self):
        super().initialize()
        self.file.truncate(0)
        self.file.seek(0)
        return self

    def eval_stmt_fallback(
        self, frame: EmitStrFrame, stmt: ir.Statement
    ) -> tuple[str, ...]:
        return (stmt.name,)

    def emit_block(self, frame: EmitStrFrame, block: ir.Block) -> str | None:
        for stmt in block.stmts:
            result = self.eval_stmt(frame, stmt)
            if isinstance(result, tuple):
                frame.set_values(stmt.results, result)
        return None

    def get_output(self) -> str:
        self.file.seek(0)
        return "\n".join(
            [
                "item    \tamount \t  price",
                "-----------------------------------",
                self.file.read(),
            ]
        )


@func.dialect.register(key="emit.recept")
class FuncEmit(interp.MethodTable):

    @interp.impl(func.Function)
    def emit_func(self, emit: EmitReceptMain, frame: EmitStrFrame, stmt: func.Function):
        _ = emit.run_ssacfg_region(frame, stmt.body)
        return ()


@dialect.register(key="emit.recept")
class FoodEmit(interp.MethodTable):

    @interp.impl(stmts.Cook)
    def emit_cook(self, emit: EmitReceptMain, frame: EmitStrFrame, stmt: stmts.Cook):
        serving_item = cast(ItemServing, emit.recept_analysis_result[stmt.result])

        amount_str = ""
        price_str = ""
        if isinstance(serving_item.count, AtLeastXItem):
            amount_str = f">={serving_item.count.data}"
            price_str = (
                f"  >=${emit.menu_price[serving_item.type] * serving_item.count.data}"
            )
        elif isinstance(serving_item.count, ConstIntItem):
            amount_str = f"  {serving_item.count.data}"
            price_str = (
                f"  ${emit.menu_price[serving_item.type] * serving_item.count.data}"
            )
        else:
            raise EmitError("invalid analysis result.")

        emit.writeln(frame, f"{serving_item.type}\t{amount_str}\t{price_str}")

        return ()



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar
from dataclasses import dataclass

from kirin.ir.traits.abc import StmtTrait

if TYPE_CHECKING:
    from kirin.ir import Region, Statement
    from kirin.dialects.func.attrs import Signature

StmtType = TypeVar("StmtType", bound="Statement")


@dataclass(frozen=True)
class CallableStmtInterface(StmtTrait, Generic[StmtType]):
    """A trait that indicates that a statement is a callable statement.

    A callable statement is a statement that can be called as a function.
    """

    @classmethod
    @abstractmethod
    def get_callable_region(cls, stmt: "StmtType") -> "Region":
        """Returns the body of the callable region"""
        ...


@dataclass(frozen=True)
class HasSignature(StmtTrait, ABC):
    """A trait that indicates that a statement has a function signature
    attribute.
    """

    @classmethod
    def get_signature(cls, stmt: "Statement"):
        signature: Signature | None = stmt.attributes.get("signature")  # type: ignore
        if signature is None:
            raise ValueError(f"Statement {stmt.name} does not have a function type")

        return signature

    @classmethod
    def set_signature(cls, stmt: "Statement", signature: "Signature"):
        stmt.attributes["signature"] = signature

    def verify(self, stmt: "Statement"):
        from kirin.dialects.func.attrs import Signature

        signature = self.get_signature(stmt)
        if not isinstance(signature, Signature):
            raise ValueError(f"{signature} is not a Signature attribute")



"""builtin dialect for python builtins

This dialect provides implementations for builtin functions like abs and sum.

- Statements: `Abs`, `Sum`.
- The lowering pass for the builtin functions.
- The concrete implementation of the builtin functions.
- The type inference implementation of the builtin functions.

This dialect maps `ast.Call` nodes of builtin functions to the `Abs` and `Sum` statements.
"""

from ast import Call

from kirin import ir, types, interp, lowering
from kirin.decl import info, statement

dialect = ir.Dialect("py.builtin")

T = types.TypeVar("T", bound=types.Int | types.Float)


@statement(dialect=dialect)
class Abs(ir.Statement):
    name = "abs"
    traits = frozenset({ir.Pure(), ir.FromPythonCall()})
    value: ir.SSAValue = info.argument(T, print=False)
    result: ir.ResultValue = info.result(T)


@statement(dialect=dialect)
class Sum(ir.Statement):
    name = "sum"
    traits = frozenset({ir.Pure(), ir.FromPythonCall()})
    value: ir.SSAValue = info.argument(types.Any, print=False)
    result: ir.ResultValue = info.result(types.Any)


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_Call_abs(
        self, state: lowering.LoweringState, node: Call
    ) -> lowering.Result:
        return lowering.Result(
            state.append_stmt(Abs(state.visit(node.args[0]).expect_one()))
        )

    def lower_Call_sum(
        self, state: lowering.LoweringState, node: Call
    ) -> lowering.Result:
        return lowering.Result(
            state.append_stmt(Sum(state.visit(node.args[0]).expect_one()))
        )


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(Abs)
    def abs(self, interp, frame: interp.Frame, stmt: Abs):
        return (abs(frame.get(stmt.value)),)

    @interp.impl(Sum)
    def _sum(self, interp, frame: interp.Frame, stmt: Sum):
        return (sum(frame.get(stmt.value)),)


@dialect.register(key="typeinfer")
class TypeInfer(interp.MethodTable):

    @interp.impl(Abs, types.Int)
    def absi(self, interp, frame, stmt):
        return (types.Int,)

    @interp.impl(Abs, types.Float)
    def absf(self, interp, frame, stmt):
        return (types.Float,)



from typing import TypeVar, Callable

from typing_extensions import Unpack, dataclass_transform

from kirin.ir import Statement
from kirin.decl import info
from kirin.decl.base import StatementOptions
from kirin.decl.verify import Verify
from kirin.decl.emit.init import EmitInit
from kirin.decl.emit.name import EmitName
from kirin.decl.emit.repr import EmitRepr
from kirin.decl.emit.traits import EmitTraits
from kirin.decl.emit.verify import EmitVerify
from kirin.decl.scan_fields import ScanFields
from kirin.decl.emit.dialect import EmitDialect
from kirin.decl.emit.property import EmitProperty
from kirin.decl.emit.typecheck import EmitTypeCheck


class StatementDecl(
    ScanFields,
    Verify,
    EmitInit,
    EmitProperty,
    EmitDialect,
    EmitName,
    EmitRepr,
    EmitTraits,
    EmitVerify,
    EmitTypeCheck,
):
    pass


StmtType = TypeVar("StmtType", bound=Statement)


@dataclass_transform(
    field_specifiers=(
        info.attribute,
        info.argument,
        info.region,
        info.result,
        info.block,
    )
)
def statement(
    cls=None,
    **kwargs: Unpack[StatementOptions],
) -> Callable[[type[StmtType]], type[StmtType]]:
    """Declare a new statement class.

    This decorator is used to declare a new statement class. It is used to
    generate the necessary boilerplate code for the class. The class should
    inherit from `kirin.ir.Statement`.

    Args:
        init(bool): Whether to generate an `__init__` method.
        repr(bool): Whether to generate a `__repr__` method.
        kw_only(bool): Whether to use keyword-only arguments in the `__init__`
            method.
        dialect(Optional[Dialect]): The dialect of the statement.
        property(bool): Whether to generate property methods for attributes.

    Example:
        The following is an example of how to use the `statement` decorator.

        ```python
        @statement
        class MyStatement(ir.Statement):
            name = "some_name"
            traits = frozenset({TraitA(), TraitB()})
            some_input: ir.SSAValue = info.argument()
            some_output: ir.ResultValue = info.result()
            body: ir.Region = info.region()
            successor: ir.Block = info.block()
        ```

        If the `name` field is not specified, a lowercase name field will be auto generated.

        In addition, one can optionally register the statement to a dialect
        by providing the `dialect` argument to the decorator.

        The following example register the statement to a dialect `my_dialect_object`, and `name = "myfoo"` field is autogenerated

        ```python
        @statement(dialect=my_dialect_object)
        class MyFoo(ir.Statement):
            traits = frozenset({ir.FromPythonCall()})
            value: str = info.attribute()
        ```
    """

    def wrap(cls):
        decl = StatementDecl(cls, **kwargs)
        decl.scan_fields()
        decl.verify()
        decl.emit()
        decl.register()
        return cls

    if cls is None:
        return wrap
    return wrap(cls)


def fields(cls: type[Statement] | Statement) -> info.StatementFields:
    return getattr(cls, ScanFields._FIELDS)



from kirin import types, interp

from . import stmts
from ._dialect import dialect


@dialect.register(key="typeinfer")
class TypeInfer(interp.MethodTable):

    @interp.impl(stmts.Add, types.Float, types.Float)
    @interp.impl(stmts.Add, types.Float, types.Int)
    @interp.impl(stmts.Add, types.Int, types.Float)
    def addf(self, interp, frame, stmt):
        return (types.Float,)

    @interp.impl(stmts.Add, types.Int, types.Int)
    def addi(self, interp, frame, stmt):
        return (types.Int,)

    @interp.impl(stmts.Sub, types.Float, types.Float)
    @interp.impl(stmts.Sub, types.Float, types.Int)
    @interp.impl(stmts.Sub, types.Int, types.Float)
    def subf(self, *_):
        return (types.Float,)

    @interp.impl(stmts.Sub, types.Int, types.Int)
    def subi(self, *_):
        return (types.Int,)

    @interp.impl(stmts.Mult, types.Float, types.Float)
    @interp.impl(stmts.Mult, types.Float, types.Int)
    @interp.impl(stmts.Mult, types.Int, types.Float)
    def multf(self, *_):
        return (types.Float,)

    @interp.impl(stmts.Mult, types.Int, types.Int)
    def multi(self, *_):
        return (types.Int,)

    @interp.impl(stmts.Div)
    def divf(self, typeinfer_, frame, stmt):
        return (types.Float,)

    @interp.impl(stmts.Mod, types.Float, types.Float)
    @interp.impl(stmts.Mod, types.Float, types.Int)
    @interp.impl(stmts.Mod, types.Int, types.Float)
    def modf(self, *_):
        return (types.Float,)

    @interp.impl(stmts.Mod, types.Int, types.Int)
    def modi(self, *_):
        return (types.Int,)

    @interp.impl(stmts.BitAnd, types.Int, types.Int)
    def bit_andi(self, interp, frame, stmt):
        return (types.Int,)

    @interp.impl(stmts.BitAnd, types.Bool, types.Bool)
    def bit_andb(self, interp, frame, stmt):
        return (types.Bool,)

    @interp.impl(stmts.BitOr, types.Int, types.Int)
    def bit_ori(self, interp, frame, stmt):
        return (types.Int,)

    @interp.impl(stmts.BitOr, types.Bool, types.Bool)
    def bit_orb(self, interp, frame, stmt):
        return (types.Bool,)

    @interp.impl(stmts.BitXor, types.Int, types.Int)
    def bit_xori(self, interp, frame, stmt):
        return (types.Int,)

    @interp.impl(stmts.BitXor, types.Bool, types.Bool)
    def bit_xorb(self, interp, frame, stmt):
        return (types.Bool,)

    @interp.impl(stmts.LShift, types.Int)
    def lshift(self, interp, frame, stmt):
        return (types.Int,)

    @interp.impl(stmts.RShift, types.Int)
    def rshift(self, interp, frame, stmt):
        return (types.Int,)

    @interp.impl(stmts.FloorDiv, types.Float, types.Float)
    @interp.impl(stmts.FloorDiv, types.Int, types.Float)
    @interp.impl(stmts.FloorDiv, types.Float, types.Int)
    def floor_divf(self, interp, frame, stmt):
        return (types.Float,)

    @interp.impl(stmts.FloorDiv, types.Int, types.Int)
    def floor_divi(self, interp, frame, stmt):
        return (types.Int,)

    @interp.impl(stmts.Pow, types.Float, types.Float)
    @interp.impl(stmts.Pow, types.Float, types.Int)
    @interp.impl(stmts.Pow, types.Int, types.Float)
    def powf(self, interp, frame, stmt):
        return (types.Float,)

    @interp.impl(stmts.Pow, types.Int, types.Int)
    def powi(self, interp, frame, stmt):
        return (types.Int,)

    @interp.impl(stmts.MatMult)
    def mat_mult(self, interp, frame, stmt):
        raise NotImplementedError("np.array @ np.array not implemented")



from typing import Generic, TypeVar
from contextlib import contextmanager
from dataclasses import field, dataclass

from kirin.interp.frame import FrameABC

FrameType = TypeVar("FrameType", bound=FrameABC)


@dataclass
class InterpreterState(Generic[FrameType]):
    """Interpreter state.

    This class represents the state of the interpreter. It contains the
    stack of frames for the interpreter. The stack of frames is used to
    store the current state of the interpreter during interpretation.
    """

    frames: list[FrameType] = field(default_factory=list)

    def push_frame(self, frame: FrameType):
        """Push a frame onto the stack.

        Args:
            frame(FrameType): The frame to push onto the stack.

        Returns:
            FrameType: The frame that was pushed.
        """
        self.frames.append(frame)
        return frame

    def pop_frame(self) -> FrameType:
        """Pop a frame from the stack.

        Returns:
            FrameType: The frame that was popped.
        """
        return self.frames.pop()

    @contextmanager
    def new_frame(self, frame: FrameType):
        """Context manager to push and pop a frame.

        This context manager is used to push a frame onto the stack before
        executing a block of code and pop the frame from the stack after
        executing the block of code.

        Args:
            frame(FrameType): The frame to push onto the stack.

        Yields:
            FrameType: The frame that was pushed.
        """
        self.push_frame(frame)
        try:
            yield frame
        finally:
            self.pop_frame()

    def current_frame(self) -> FrameType:
        """Get the current frame.

        Returns:
            FrameType: The current frame.
        """
        if not self.frames:
            raise ValueError(
                "unable to retrieve the current frame because no frames were pushed"
            )
        return self.frames[-1]



from kirin import ir
from kirin.decl import info
from kirin.decl.emit.init import BaseModifier

from ._create_fn import create_fn
from ._set_new_attribute import set_new_attribute


class EmitProperty(BaseModifier):
    _KIRIN_PYATTR = "_kirin_PyAttr"

    def emit_property(self):
        for name, f in self.fields.args.items():
            getter, setter = self._emit_arg_property(f)
            set_new_attribute(self.cls, name, property(getter, setter))

        for name, f in self.fields.kw_args.items():
            getter, setter = self._emit_arg_property(f)
            set_new_attribute(self.cls, name, property(getter, setter))

        for i, (name, f) in enumerate(self.fields.results.items()):
            getter, setter = self._emit_result_property(i, f)
            set_new_attribute(self.cls, name, property(getter, setter))

        for name, f in self.fields.attributes.items():
            getter, setter = self._emit_attribute_property(f)
            set_new_attribute(self.cls, name, property(getter, setter))

        for i, (name, f) in enumerate(self.fields.blocks.items()):
            getter, setter = self._emit_successor_property(i, f)
            set_new_attribute(self.cls, name, property(getter, setter))

        for i, (name, f) in enumerate(self.fields.regions.items()):
            getter, setter = self._emit_region_property(i, f)
            set_new_attribute(self.cls, name, property(getter, setter))

    def _emit_arg_property(self, f: info.ArgumentField):
        getter = create_fn(
            f"_get_{f.name}",
            args=[self._self_name],
            body=[
                f'return {self._self_name}._args[{self._self_name}._name_args_slice["{f.name}"]]'
            ],
            globals=self.globals,
            return_type=tuple[ir.SSAValue, ...] if f.group else ir.SSAValue,
        )

        if f.group:
            return getter, self._emit_arg_property_group_setter(f)
        return getter, self._emit_arg_property_std_setter(f)

    def _emit_arg_property_std_setter(self, f: info.ArgumentField):
        return create_fn(
            f"_set_{f.name}",
            args=[self._self_name, "value: _value_hint"],
            body=[
                f"s = {self._self_name}._name_args_slice['{f.name}']",
                f"old = {self._self_name}.args[s]",
                f"old.remove_use(Use({self._self_name}, s))",
                f"value.add_use(Use({self._self_name}, s))",
                "self.args = (*self.args[:s], value, *self.args[s + 1:])",
            ],
            globals=self.globals,
            locals={"_value_hint": ir.SSAValue},
            return_type=None,
        )

    def _emit_arg_property_group_setter(self, f: info.ArgumentField):
        return create_fn(
            f"_set_{f.name}",
            args=[self._self_name, "value: _value_hint"],
            body=[
                f"s = {self._self_name}._name_args_slice['{f.name}']",
                "start, stop = s.start, s.stop",
                f"old = {self._self_name}.args[s]",
                "a_range = range(start, stop, s.step or 1)",
                "for i, arg in zip(a_range, old):",
                f"    arg.remove_use(Use({self._self_name}, i))",
                "for i, arg in zip(a_range, value):",
                f"    arg.add_use(Use({self._self_name}, i))",
                f"{self._self_name}.args = (*{self._self_name}.args[:start], *value, *{self._self_name}.args[stop + 1:])",
            ],
            globals=self.globals,
            locals={"_value_hint": tuple[ir.SSAValue, ...]},
            return_type=None,
        )

    def _emit_result_property(self, index: int, f: info.ResultField):
        getter = create_fn(
            f"_get_{f.name}",
            args=[self._self_name],
            body=[f"return {self._self_name}._results[{index}]"],
            globals=self.globals,
            return_type=ir.ResultValue,
        )

        # well you cannot delete what already happened, can you?
        setter = create_fn(
            f"_set_{f.name}",
            args=[self._self_name, "value: _value_hint"],
            body=[f"raise AttributeError('result property {f.name} is read-only')"],
            globals=self.globals,
            locals={"_value_hint": ir.ResultValue},
            return_type=None,
        )

        return getter, setter

    def _emit_attribute_property(self, f: info.AttributeField):
        from kirin.ir.attrs.py import PyAttr

        storage = "attributes"
        attr = f"{self._self_name}.{storage}['{f.name}']"
        getter = create_fn(
            f"_get_{f.name}",
            args=[self._self_name],
            body=[f"return {attr}.data" if f.pytype else f"return {attr}"],
            globals=self.globals,
            return_type=f.annotation,
        )

        setter = create_fn(
            f"_set_{f.name}",
            args=[self._self_name, "value: _value_hint"],
            body=[
                (
                    f"{attr} = value if isinstance(value, {self._KIRIN_PYATTR}) else {self._KIRIN_PYATTR}(value)"
                    if f.pytype
                    else f"{attr} = value"
                )
            ],
            globals=self.globals,
            locals={"_value_hint": PyAttr if f.pytype else f.annotation},
            return_type=None,
        )

        return getter, setter

    def _emit_successor_property(self, index: int, f: info.BlockField):
        getter = create_fn(
            f"_get_{f.name}",
            args=[self._self_name],
            body=[f"return {self._self_name}.successors[{index}]"],
            globals=self.globals,
            return_type=ir.Block,
        )

        setter = create_fn(
            f"_set_{f.name}",
            args=[self._self_name, "value: _value_hint"],
            body=[f"{self._self_name}.successors[{index}] = value"],
            globals=self.globals,
            locals={"_value_hint": ir.Block},
            return_type=None,
        )

        return getter, setter

    def _emit_region_property(self, index: int, f: info.RegionField):
        getter = create_fn(
            f"_get_{f.name}",
            args=[self._self_name],
            body=[f"return {self._self_name}.regions[{index}]"],
            globals=self.globals,
            return_type=ir.Region,
        )

        setter = create_fn(
            f"_set_{f.name}",
            args=[self._self_name, "value: _value_hint"],
            body=[
                f"old = {self._self_name}.regions[{index}]",
                "old.parent = None",
                f"value.parent = {self._self_name}",
                f"{self._self_name}.regions[{index}] = value",
            ],
            globals=self.globals,
            locals={"_value_hint": ir.Region},
            return_type=None,
        )

        return getter, setter



# type: ignore
from kirin.prelude import python_no_opt
from kirin.dialects import py


def test_list_append():

    @python_no_opt
    def test_append():
        x = []
        py.Append(x, 1)
        py.Append(x, 2)
        return x

    y = test_append()

    assert len(y) == 2
    assert y[0] == 1
    assert y[1] == 2


def test_recursive_append():
    @python_no_opt
    def for_loop_append(cntr: int, x: list, n_range: int):
        if cntr < n_range:
            py.Append(x, cntr)
            for_loop_append(cntr + 1, x, n_range)

        return x

    assert for_loop_append(0, [], 5) == [0, 1, 2, 3, 4]



"""The `Len` dialect.

This dialect maps the `len()` call to the `Len` statement:

- The `Len` statement class.
- The lowering pass for the `len()` call.
- The concrete implementation of the `len()` call.
"""

import ast

from kirin import ir, types, interp, lowering
from kirin.decl import info, statement
from kirin.analysis import const

dialect = ir.Dialect("py.len")


@statement(dialect=dialect)
class Len(ir.Statement):
    name = "len"
    traits = frozenset({ir.Pure(), ir.FromPythonCall()})
    value: ir.SSAValue = info.argument(types.Any)
    result: ir.ResultValue = info.result(types.Int)


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(Len)
    def len(self, interp, frame: interp.Frame, stmt: Len):
        return (len(frame.get(stmt.value)),)


@dialect.register(key="constprop")
class ConstProp(interp.MethodTable):

    @interp.impl(Len)
    def len(self, interp, frame: interp.Frame, stmt: Len):
        value = frame.get(stmt.value)
        if isinstance(value, const.Value):
            return (const.Value(len(value.data)),)
        elif isinstance(value, const.PartialTuple):
            return (const.Value(len(value.data)),)
        else:
            return (const.Result.top(),)


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_Call_len(
        self, state: lowering.LoweringState, node: ast.Call
    ) -> lowering.Result:
        return lowering.Result(
            state.append_stmt(Len(value=state.visit(node.args[0]).expect_one()))
        )



"""Built-in dialects for Kirin.

This module contains the built-in dialects for Kirin. Each dialect is an
instance of the `Dialect` class. Each submodule contains a `dialect` variable
that is an instance of the corresponding `Dialect` class.

The modules can be directly used as dialects. For example, you can write

```python
from kirin.dialects import py, func
```

to import the Python and function dialects.
"""



import ast
from typing import TYPE_CHECKING, TypeVar
from dataclasses import dataclass

from kirin.exceptions import DialectLoweringError

from ..abc import PythonLoweringTrait

if TYPE_CHECKING:
    from kirin.ir import Statement
    from kirin.lowering import Result, LoweringState


StatementType = TypeVar("StatementType", bound="Statement")


@dataclass(frozen=True)
class FromPythonCall(PythonLoweringTrait[StatementType, ast.Call]):
    """Trait for customizing lowering of Python calls to a statement.

    Declared in a statement definition to indicate that the statement can be
    constructed from a Python call (i.e., a function call `ast.Call` in the
    Python AST).

    Subclassing this trait allows for customizing the lowering of Python calls
    to the statement. The `lower` method should be implemented to parse the
    arguments from the Python call and construct the statement instance.
    """

    def lower(
        self, stmt: type[StatementType], state: "LoweringState", node: ast.Call
    ) -> "Result":
        return state.default_Call_lower(stmt, node)

    def verify(self, stmt: "Statement"):
        assert len(stmt.regions) == 0, "FromPythonCall statements cannot have regions"
        assert (
            len(stmt.successors) == 0
        ), "FromPythonCall statements cannot have successors"


@dataclass(frozen=True)
class FromPythonRangeLike(FromPythonCall[StatementType]):
    """Provides a default lowering implementation for built-in `range`-like function
    to a statement that takes three arguments: start, stop, and step.
    """

    def lower(
        self, stmt: type[StatementType], state: "LoweringState", node: ast.Call
    ) -> "Result":
        from kirin.lowering import Result

        if len(node.args) == 1:
            start = state.visit(ast.Constant(0)).expect_one()
            stop = state.visit(node.args[0]).expect_one()
            step = state.visit(ast.Constant(1)).expect_one()
        elif len(node.args) == 2:
            start = state.visit(node.args[0]).expect_one()
            stop = state.visit(node.args[1]).expect_one()
            step = state.visit(ast.Constant(1)).expect_one()
        elif len(node.args) == 3:
            start = state.visit(node.args[0]).expect_one()
            stop = state.visit(node.args[1]).expect_one()
            step = state.visit(node.args[2]).expect_one()
        else:
            raise DialectLoweringError("range() takes 1-3 arguments")

        return Result(state.append_stmt(stmt(start, stop, step)))  # type: ignore



"""Assertion dialect for Python.

This module contains the dialect for the Python `assert` statement, including:

- The `Assert` statement class.
- The lowering pass for the `assert` statement.
- The concrete implementation of the `assert` statement.
- The type inference implementation of the `assert` statement.
- The Julia emitter for the `assert` statement.

This dialect maps `ast.Assert` nodes to the `Assert` statement.
"""

import ast

from kirin import ir, types, interp, lowering
from kirin.decl import info, statement
from kirin.emit import EmitStrFrame, julia
from kirin.print import Printer

dialect = ir.Dialect("py.assert")


@statement(dialect=dialect)
class Assert(ir.Statement):
    condition: ir.SSAValue
    message: ir.SSAValue = info.argument(types.String)

    def print_impl(self, printer: Printer) -> None:
        with printer.rich(style="keyword"):
            printer.print_name(self)

        printer.plain_print(" ")
        printer.print(self.condition)

        if self.message:
            printer.plain_print(", ")
            printer.print(self.message)


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_Assert(
        self, state: lowering.LoweringState, node: ast.Assert
    ) -> lowering.Result:
        from kirin.dialects.py.constant import Constant

        cond = state.visit(node.test).expect_one()
        if node.msg:
            message = state.visit(node.msg).expect_one()
            state.append_stmt(Assert(condition=cond, message=message))
        else:
            message_stmt = state.append_stmt(Constant(""))
            state.append_stmt(Assert(condition=cond, message=message_stmt.result))
        return lowering.Result()


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(Assert)
    def assert_stmt(
        self, interp_: interp.Interpreter, frame: interp.Frame, stmt: Assert
    ):
        if frame.get(stmt.condition) is True:
            return ()

        if stmt.message:
            raise interp.WrapException(AssertionError(frame.get(stmt.message)))
        else:
            raise interp.WrapException(AssertionError("Assertion failed"))


@dialect.register(key="typeinfer")
class TypeInfer(interp.MethodTable):

    @interp.impl(Assert)
    def assert_stmt(self, interp, frame, stmt: Assert):
        return (types.Bottom,)


@dialect.register(key="emit.julia")
class EmitJulia(interp.MethodTable):

    @interp.impl(Assert)
    def emit_assert(self, interp: julia.EmitJulia, frame: EmitStrFrame, stmt: Assert):
        interp.writeln(
            frame, f"@assert {frame.get(stmt.condition)} {frame.get(stmt.message)}"
        )
        return ()



from typing import Callable
from dataclasses import field, dataclass

from kirin.ir import Block, Region, Statement
from kirin.worklist import WorkList
from kirin.rewrite.abc import RewriteRule
from kirin.ir.nodes.base import IRNode
from kirin.rewrite.result import RewriteResult


@dataclass
class Walk(RewriteRule):
    """Walk through the IR nodes and apply a rewrite rule.

    The walk will apply the rewrite rule to each node in the IR tree in a depth-first order.

    ### Parameters
    - `map`: The rewrite rule to apply.
    - `reverse`: Whether to traverse the IR tree in reverse order. Default is `False`.
    - `region_first`: Whether to traverse the regions before the blocks. Default is `False`.
    """

    rule: RewriteRule
    worklist: WorkList[IRNode] = field(default_factory=WorkList)
    skip: Callable[[IRNode], bool] = field(default=lambda _: False)

    # options
    reverse: bool = field(default=False)
    region_first: bool = field(default=False)

    def rewrite(self, node: IRNode) -> RewriteResult:
        # NOTE: because the rewrite pass may mutate the node
        # thus we need to save the list of nodes to be processed
        # first before we start processing them
        self.populate_worklist(node)
        has_done_something = False
        subnode = self.worklist.pop()
        while subnode is not None:
            result = self.rule.rewrite(subnode)
            if result.terminated:
                return result

            if result.has_done_something:
                has_done_something = True
            subnode = self.worklist.pop()
        return RewriteResult(has_done_something=has_done_something)

    def populate_worklist(self, node: IRNode) -> None:
        if self.skip(node):
            return

        if isinstance(node, Statement):
            self.populate_worklist_Statement(node)
        elif isinstance(node, Region):
            self.populate_worklist_Region(node)
        elif isinstance(node, Block):
            self.populate_worklist_Block(node)
        else:
            raise NotImplementedError(f"populate_worklist_{node.__class__.__name__}")

    def populate_worklist_Statement(self, node: Statement) -> None:
        if self.region_first:
            self.worklist.append(node)

        if node.regions:
            for region in reversed(node.regions) if not self.reverse else node.regions:
                self.populate_worklist(region)

        if not self.region_first:
            self.worklist.append(node)

    def populate_worklist_Region(self, node: Region) -> None:
        self.worklist.append(node)
        if node.blocks:
            for block in reversed(node.blocks) if not self.reverse else node.blocks:
                self.populate_worklist(block)

    def populate_worklist_Block(self, node: Block) -> None:
        self.worklist.append(node)
        stmt = node.last_stmt
        while stmt is not None:
            self.populate_worklist(stmt)
            stmt = stmt.prev_stmt



from typing import Generic, TypeVar, Callable, Iterable, TypeAlias
from dataclasses import dataclass

from kirin.ir.group import DialectGroup
from kirin.ir.nodes import Statement
from kirin.lowering import FromPythonAST
from kirin.interp.base import FrameABC, BaseInterpreter
from kirin.interp.impl import Signature
from kirin.interp.table import MethodTable
from kirin.interp.value import StatementResult
from kirin.ir.attrs.abc import Attribute

MethodTableSelf = TypeVar("MethodTableSelf", bound="MethodTable")
InterpreterType = TypeVar("InterpreterType", bound="BaseInterpreter")
FrameType = TypeVar("FrameType", bound="FrameABC")
StatementType = TypeVar("StatementType", bound="Statement")
MethodFunction: TypeAlias = Callable[
    [MethodTableSelf, InterpreterType, FrameType, StatementType], StatementResult
]

@dataclass
class StatementImpl(Generic[InterpreterType, FrameType]):
    parent: "MethodTable"
    impl: MethodFunction["MethodTable", InterpreterType, FrameType, "Statement"]

    def __call__(
        self, interp: InterpreterType, frame: FrameType, stmt: "Statement"
    ) -> StatementResult: ...
    def __repr__(self) -> str: ...

@dataclass
class AttributeImpl:
    parent: "MethodTable"
    impl: Callable

    def __call__(self, interp, attr: "Attribute"): ...
    def __repr__(self) -> str: ...

@dataclass
class InterpreterRegistry:
    attributes: dict[type["Attribute"], "AttributeImpl"]
    statements: dict["Signature", "StatementImpl"]

@dataclass
class Registry:
    """Proxy class to build different registries from a dialect group."""

    dialects: "DialectGroup"
    """The dialect group to build the registry from."""

    def ast(self, keys: Iterable[str]) -> dict[str, "FromPythonAST"]: ...
    def interpreter(self, keys: Iterable[str]) -> InterpreterRegistry: ...



import ast

from kirin import lowering, exceptions

from . import stmts
from ._dialect import dialect


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_UnaryOp(
        self, state: lowering.LoweringState, node: ast.UnaryOp
    ) -> lowering.Result:
        if op := getattr(stmts, node.op.__class__.__name__, None):
            return lowering.Result(
                state.append_stmt(op(state.visit(node.operand).expect_one()))
            )
        else:
            raise exceptions.DialectLoweringError(
                f"unsupported unary operator {node.op}"
            )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar, Iterator
from dataclasses import dataclass

from typing_extensions import Self

from kirin.print import Printer, Printable
from kirin.ir.ssa import SSAValue

if TYPE_CHECKING:
    from kirin.ir.nodes.stmt import Statement


ParentType = TypeVar("ParentType", bound="IRNode")


@dataclass
class IRNode(Generic[ParentType], ABC, Printable):
    """Base class for all IR nodes. All IR nodes are hashable and can be compared
    for equality. The hash of an IR node is the same as the id of the object.

    !!! note "Pretty Printing"
        This object is pretty printable via
        [`.print()`][kirin.print.printable.Printable.print] method.
    """

    def assert_parent(self, type_: type[IRNode], parent) -> None:
        assert (
            isinstance(parent, type_) or parent is None
        ), f"Invalid parent, expect {type_} or None, got {type(parent)}"

    @property
    @abstractmethod
    def parent_node(self) -> ParentType | None:
        """Parent node of the current node."""
        ...

    @parent_node.setter
    @abstractmethod
    def parent_node(self, parent: ParentType | None) -> None: ...

    def is_ancestor(self, op: IRNode) -> bool:
        """Check if the given node is an ancestor of the current node."""
        if op is self:
            return True
        if (parent := op.parent_node) is None:
            return False
        return self.is_ancestor(parent)

    def get_root(self) -> IRNode:
        """Get the root node of the current node."""
        if (parent := self.parent_node) is None:
            return self
        return parent.get_root()

    def is_equal(self, other: IRNode, context: dict = {}) -> bool:
        """Check if the current node is equal to the other node.

        Args:
            other: The other node to compare.
            context: The context to store the visited nodes. Defaults to {}.

        Returns:
            True if the nodes are equal, False otherwise.

        !!! note
            This method is not the same as the `==` operator. It checks for
            structural equality rather than identity. To change the behavior
            of structural equality, override the `is_structurally_equal` method.
        """
        if not isinstance(other, type(self)):
            return False
        return self.is_structurally_equal(other, context)

    def attach(self, parent: ParentType) -> None:
        """Attach the current node to the parent node."""
        assert isinstance(parent, IRNode), f"Expected IRNode, got {type(parent)}"

        if self.parent_node:
            raise ValueError("Node already has a parent")
        if self.is_ancestor(parent):
            raise ValueError("Node is an ancestor of the parent")
        self.parent_node = parent

    @abstractmethod
    def detach(self) -> None:
        """Detach the current node from the parent node."""
        ...

    @abstractmethod
    def drop_all_references(self) -> None:
        """Drop all references to other nodes."""
        ...

    @abstractmethod
    def delete(self, safe: bool = True) -> None:
        """Delete the current node.

        Args:
            safe: If True, check if the node has any references before deleting.
        """
        ...

    @abstractmethod
    def is_structurally_equal(
        self,
        other: Self,
        context: dict[IRNode | SSAValue, IRNode | SSAValue] | None = None,
    ) -> bool:
        """Check if the current node is structurally equal to the other node.

        !!! note
            This method is for tweaking the behavior of structural equality.
            To check if two nodes are structurally equal, use the `is_equal` method.

        Args:
            other: The other node to compare.
            context: The context to store the visited nodes.

        Returns:
            True if the nodes are structurally equal, False otherwise.
        """
        ...

    def __eq__(self, other) -> bool:
        return self is other

    def __hash__(self) -> int:
        return id(self)

    @abstractmethod
    def walk(
        self, *, reverse: bool = False, region_first: bool = False
    ) -> Iterator[Statement]: ...

    @abstractmethod
    def print_impl(self, printer: Printer) -> None: ...

    @abstractmethod
    def typecheck(self) -> None:
        """check if types are correct."""
        ...

    @abstractmethod
    def verify(self) -> None:
        """run mandatory validation checks. This is not same as typecheck, which may be optional."""
        ...



import ast

from kirin import types
from kirin.lowering import Result, FromPythonAST, LoweringState

from . import stmts as ilist
from ._dialect import dialect


@dialect.register
class PythonLowering(FromPythonAST):

    def lower_List(self, state: LoweringState, node: ast.List) -> Result:
        elts = tuple(state.visit(each).expect_one() for each in node.elts)

        if len(elts):
            typ = elts[0].type
            for each in elts:
                typ = typ.join(each.type)
        else:
            typ = types.Any

        stmt = ilist.New(values=tuple(elts))
        state.append_stmt(stmt)
        return Result(stmt)



import ast

from kirin import ir, types, lowering
from kirin.dialects import func
from kirin.exceptions import DialectLoweringError

dialect = ir.Dialect("lowering.call")


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_Call_local(
        self, state: lowering.LoweringState, callee: ir.SSAValue, node: ast.Call
    ) -> lowering.Result:
        args, keywords = self.__lower_Call_args_kwargs(state, node)
        stmt = func.Call(callee, args, kwargs=keywords)
        return lowering.Result(state.append_stmt(stmt))

    def lower_Call_global_method(
        self,
        state: lowering.LoweringState,
        method: ir.Method,
        node: ast.Call,
    ) -> lowering.Result:
        args, keywords = self.__lower_Call_args_kwargs(state, node)
        stmt = func.Invoke(args, callee=method, kwargs=keywords)
        stmt.result.type = method.return_type or types.Any
        return lowering.Result(state.append_stmt(stmt))

    def __lower_Call_args_kwargs(
        self,
        state: lowering.LoweringState,
        node: ast.Call,
    ):
        args: list[ir.SSAValue] = []
        for arg in node.args:
            if isinstance(arg, ast.Starred):  # TODO: support *args
                raise DialectLoweringError("starred arguments are not supported")
            else:
                args.append(state.visit(arg).expect_one())

        keywords = []
        for kw in node.keywords:
            keywords.append(kw.arg)
            args.append(state.visit(kw.value).expect_one())

        return tuple(args), tuple(keywords)



import ast

from kirin import lowering, exceptions

from . import stmts
from ._dialect import dialect


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_BinOp(
        self, state: lowering.LoweringState, node: ast.BinOp
    ) -> lowering.Result:
        lhs = state.visit(node.left).expect_one()
        rhs = state.visit(node.right).expect_one()

        if op := getattr(stmts, node.op.__class__.__name__, None):
            stmt = op(lhs=lhs, rhs=rhs)
        else:
            raise exceptions.DialectLoweringError(f"unsupported binop {node.op}")
        return lowering.Result(state.append_stmt(stmt))



"""Const analysis module.

This module contains the constant analysis framework for kirin. The constant
analysis framework is built on top of the interpreter framework.

This module provides a lattice for constant propagation analysis and a
propagation algorithm for computing the constant values for each SSA value in
the IR.
"""

from .prop import Frame as Frame, Propagate as Propagate
from .lattice import (
    Value as Value,
    Bottom as Bottom,
    Result as Result,
    Unknown as Unknown,
    PartialConst as PartialConst,
    PartialTuple as PartialTuple,
    PartialLambda as PartialLambda,
)



from typing import Generic, TypeVar, Sequence
from dataclasses import field, dataclass

Stmt = TypeVar("Stmt")


@dataclass
class StmtStream(Generic[Stmt]):
    stmts: list[Stmt] = field(default_factory=list)
    cursor: int = 0

    def __init__(self, stmts: Sequence[Stmt], cursor: int = 0):
        self.stmts = list(stmts)
        self.cursor = cursor

    def __iter__(self):
        return self

    def __next__(self):
        if self.cursor < len(self.stmts):
            stmt = self.stmts[self.cursor]
            self.cursor += 1
            return stmt
        else:
            raise StopIteration

    def peek(self):
        return self.stmts[self.cursor]

    def has_next(self):
        return self.cursor < len(self.stmts)

    def split(self) -> "StmtStream":
        cursor = self.cursor
        self.cursor = len(self.stmts)
        return StmtStream(self.stmts, cursor)

    def __len__(self):
        return len(self.stmts)

    def __getitem__(self, key):
        return self.stmts[key]

    def __setitem__(self, key, value):
        self.stmts[key] = value

    def pop(self):
        stmt = self.stmts[self.cursor]
        self.cursor += 1
        return stmt

    def is_empty(self):
        return self.cursor == len(self.stmts)



from typing import Iterable
from dataclasses import dataclass

from kirin.ir import IRNode
from kirin.rewrite.abc import RewriteRule
from kirin.rewrite.result import RewriteResult


@dataclass
class Chain(RewriteRule):
    """Chain multiple rewrites together.

    The chain will apply each rewrite in order until one of the rewrites terminates.
    """

    rules: list[RewriteRule]

    def __init__(self, rule: RewriteRule | Iterable[RewriteRule], *others: RewriteRule):
        if isinstance(rule, RewriteRule):
            self.rules = [rule, *others]
        else:
            assert (
                others == ()
            ), "Cannot pass multiple positional arguments if the first argument is an iterable"
            self.rules = list(rule)

    def rewrite(self, node: IRNode) -> RewriteResult:
        has_done_something = False
        for rule in self.rules:
            result = rule.rewrite(node)
            if result.terminated:
                return result

            if result.has_done_something:
                has_done_something = True
        return RewriteResult(has_done_something=has_done_something)

    def __repr__(self):
        return " -> ".join(map(str, self.rules))



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, Iterator
from dataclasses import field, dataclass
from collections.abc import Sequence

from typing_extensions import Self

from kirin.print import Printer
from kirin.ir.ssa import SSAValue, BlockArgument
from kirin.exceptions import VerificationError
from kirin.ir.nodes.base import IRNode
from kirin.ir.nodes.view import View, MutableSequenceView

if TYPE_CHECKING:
    from kirin.ir.nodes.stmt import Statement
    from kirin.ir.attrs.types import TypeAttribute
    from kirin.ir.nodes.region import Region


@dataclass
class BlockArguments(MutableSequenceView[tuple, "Block", BlockArgument]):
    """A View object that contains a list of BlockArgument.

    Description:
        This is a proxy object that provide safe API to manipulate the arguments of a Block.


    """

    def append_from(self, typ: TypeAttribute, name: str | None = None) -> BlockArgument:
        """Append a new argument to the Block that this View reference to.

        Description:
            This method will create a new [`BlockArgument`][kirin.ir.BlockArgument] and append it to the argument list
            of the reference `Block`.

        Args:
            typ (TypeAttribute): The type of the argument.
            name (str | None, optional): name of the argument. Defaults to `None`.

        Returns:
            BlockArgument: The newly created [`BlockArgument`][kirin.ir.BlockArgument].

        """
        new_arg = BlockArgument(self.node, len(self.node._args), typ)
        if name:
            new_arg.name = name

        self.node._args += (new_arg,)
        return new_arg

    def insert_from(
        self, idx: int, typ: TypeAttribute, name: str | None = None
    ) -> BlockArgument:
        """Insert a new argument to the Block that this View reference to.

        Description:
            This method will create a new `BlockArgument` and insert it to the argument list
            of the reference Block at the specified index

        Args:
            idx (int): Insert location index.
            typ (TypeAttribute): The type of the argument.
            name (str | None, optional): Name of the argument. Defaults to `None`.

        Returns:
            BlockArgument: The newly created BlockArgument.
        """
        if idx < 0 or idx > len(self.node._args):
            raise ValueError("Invalid index")

        new_arg = BlockArgument(self.node, idx, typ)
        if name:
            new_arg.name = name

        for arg in self.node._args[idx:]:
            arg.index += 1
        self.node._args = self.node._args[:idx] + (new_arg,) + self.node._args[idx:]
        return new_arg

    def delete(self, arg: BlockArgument, safe: bool = True) -> None:
        """Delete a BlockArgument from the Block that this View reference to.


        Args:
            arg (BlockArgument): _description_
            safe (bool, optional): If True, error will be raised if the BlockArgument has any Use by others.  Defaults to True.

        Raises:
            ValueError: If the argument does not belong to the reference block.
        """
        if arg.block is not self.node:
            raise ValueError("Attempt to delete an argument that is not in the block")

        for block_arg in self.field[arg.index + 1 :]:
            block_arg.index -= 1
        self.node._args = (*self.field[: arg.index], *self.field[arg.index + 1 :])
        arg.delete(safe=safe)

    def __delitem__(self, idx: int) -> None:
        self.delete(self.field[idx])


@dataclass
class BlockStmtIterator:
    """Proxy object to iterate over the Statements in a Block."""

    next_stmt: Statement | None

    def __iter__(self) -> BlockStmtIterator:
        return self

    def __next__(self) -> Statement:
        if self.next_stmt is None:
            raise StopIteration
        stmt = self.next_stmt
        self.next_stmt = stmt.next_stmt
        return stmt


@dataclass
class BlockStmtsReverseIterator:
    """Proxy object to iterate over the Statements in a Block in reverse order."""

    next_stmt: Statement | None

    def __iter__(self) -> BlockStmtsReverseIterator:
        return self

    def __next__(self) -> Statement:
        if self.next_stmt is None:
            raise StopIteration
        stmt = self.next_stmt
        self.next_stmt = stmt.prev_stmt
        return stmt


@dataclass
class BlockStmts(View["Block", "Statement"]):
    """A View object that contains a list of Statements.

    Description:
        This is a proxy object that provide safe API to manipulate the statements of a Block.
    """

    def __iter__(self) -> Iterator[Statement]:
        return BlockStmtIterator(self.node.first_stmt)

    def __len__(self) -> int:
        return self.node._stmt_len

    def __reversed__(self) -> Iterator[Statement]:
        return BlockStmtsReverseIterator(self.node.last_stmt)

    def __repr__(self) -> str:
        return f"BlockStmts(len={len(self)})"

    def __getitem__(self, index: int) -> Statement:
        raise NotImplementedError("Use at() instead")

    def at(self, index: int) -> Statement:
        """This is similar to __getitem__ but due to the nature of the linked list,
        it is less efficient than __getitem__.

        Args:
            index (int): Index of the Statement.

        Returns:
            Statement: The Statement at the specified index.
        """
        if index >= len(self):
            raise IndexError("Index out of range")

        if index < 0:
            return self._at_reverse(-index - 1)

        return self._at_forward(index)

    def _at_forward(self, index: int) -> Statement:
        if self.node.first_stmt is None:
            raise IndexError("Index out of range")

        stmt = self.node.first_stmt
        for _ in range(index):
            if stmt is None:
                raise IndexError("Index out of range")
            stmt = stmt.next_stmt

        if stmt is None:
            raise IndexError("Index out of range")
        return stmt

    def _at_reverse(self, index: int) -> Statement:
        if self.node.last_stmt is None:
            raise IndexError("Index out of range")

        stmt = self.node.last_stmt
        for _ in range(index):
            if stmt is None:
                raise IndexError("Index out of range")
            stmt = stmt.prev_stmt

        if stmt is None:
            raise IndexError("Index out of range")
        return stmt

    def append(self, value: Statement) -> None:
        """Append a Statement to the reference Block.

        Args:
            value (Statement): A Statement to be appended.
        """
        from kirin.ir.nodes.stmt import Statement

        if not isinstance(value, Statement):
            raise ValueError(f"Expected Statement, got {type(value).__name__}")

        if self.node._stmt_len == 0:  # empty block
            value.attach(self.node)
            self.node._first_stmt = value
            self.node._last_stmt = value
            self.node._stmt_len += 1
        elif self.node._last_stmt:
            value.insert_after(self.node._last_stmt)
        else:
            raise ValueError("Invalid block, last_stmt is None")


@dataclass
class Block(IRNode["Region"]):
    """
    Block consist of a list of Statements and optionally input arguments.

    !!! note "Pretty Printing"
        This object is pretty printable via
        [`.print()`][kirin.print.printable.Printable.print] method.
    """

    _args: tuple[BlockArgument, ...]

    # NOTE: we need linked list since stmts are inserted frequently
    _first_stmt: Statement | None = field(repr=False)
    _last_stmt: Statement | None = field(repr=False)
    _stmt_len: int = field(default=0, repr=False)

    parent: Region | None = field(default=None, repr=False)
    """Parent Region of the Block."""

    def __init__(
        self,
        stmts: Sequence[Statement] = (),
        argtypes: Iterable[TypeAttribute] = (),
    ):
        """
        Args:
            stmts (Sequence[Statement], optional): A list of statements. Defaults to ().
            argtypes (Iterable[TypeAttribute], optional): The type of the block arguments. Defaults to ().
        """
        super().__init__()
        self._args = tuple(
            BlockArgument(self, i, argtype) for i, argtype in enumerate(argtypes)
        )

        self._first_stmt = None
        self._last_stmt = None
        self._first_branch = None
        self._last_branch = None
        self._stmt_len = 0
        self.stmts.extend(stmts)

    @property
    def parent_stmt(self) -> Statement | None:
        """parent statement of the Block."""
        if self.parent is None:
            return None
        return self.parent.parent_node

    @property
    def parent_node(self) -> Region | None:
        """Get parent Region of the Block."""
        return self.parent

    @parent_node.setter
    def parent_node(self, parent: Region | None) -> None:
        """Set the parent Region of the Block."""
        from kirin.ir.nodes.region import Region

        self.assert_parent(Region, parent)
        self.parent = parent

    @property
    def args(self) -> BlockArguments:
        """Get the  arguments of the Block.

        Returns:
            BlockArguments: The arguments view of the Block.
        """
        return BlockArguments(self, self._args)

    @property
    def first_stmt(self) -> Statement | None:
        """Get the first Statement of the Block.

        Returns:
            Statement | None: The first Statement of the Block.
        """
        return self._first_stmt

    @property
    def last_stmt(self) -> Statement | None:
        """Get the last Statement of the Block.

        Returns:
            Statement | None: The last Statement of the Block.

        """
        return self._last_stmt

    @property
    def stmts(self) -> BlockStmts:
        """Get the list of Statements of the Block.

        Returns:
            BlockStmts: The Statements of the Block.
        """
        return BlockStmts(self)

    def drop_all_references(self) -> None:
        """Remove all the dependency that reference/uses this Block."""
        self.parent = None
        for stmt in self.stmts:
            stmt.drop_all_references()

    def detach(self) -> None:
        """Detach this Block from the IR.

        Note:
            Detach only detach the Block from the IR graph. It does not remove uses that reference the Block.
        """
        if self.parent is None:
            return

        idx = self.parent[self]
        del self.parent._blocks[idx]
        del self.parent._block_idx[self]
        for block in self.parent._blocks[idx:]:
            self.parent._block_idx[block] -= 1
        self.parent = None

    def delete(self, safe: bool = True) -> None:
        """Delete the Block completely from the IR.

        Note:
            This method will detach + remove references of the block.

        Args:
            safe (bool, optional): If True, raise error if there is anything that still reference components in the block. Defaults to True.
        """
        self.detach()
        self.drop_all_references()
        for stmt in self.stmts:
            stmt.delete(safe=safe)

    def is_structurally_equal(
        self,
        other: Self,
        context: dict[IRNode | SSAValue, IRNode | SSAValue] | None = None,
    ) -> bool:
        """Check if the Block is structurally equal to another Block.

        Args:
            other (Self): The other Block to compare with.
            context (dict[IRNode  |  SSAValue, IRNode  |  SSAValue] | None, optional): A map of IRNode/SSAValue to hint that they are equivalent so the check will treat them as equivalent. Defaults to None.

        Returns:
            bool: True if the Block is structurally equal to the other Block.
        """
        if context is None:
            context = {}

        if len(self._args) != len(other._args) or len(self.stmts) != len(other.stmts):
            return False

        for arg, other_arg in zip(self._args, other._args):
            if arg.type != other_arg.type:
                return False
            context[arg] = other_arg

        context[self] = other
        if not all(
            stmt.is_structurally_equal(other_stmt, context)
            for stmt, other_stmt in zip(self.stmts, other.stmts)
        ):
            return False

        return True

    def __hash__(self) -> int:
        return id(self)

    def walk(
        self, *, reverse: bool = False, region_first: bool = False
    ) -> Iterator[Statement]:
        """Traversal the Statements in a Block.

        Args:
            reverse (bool, optional): If walk in the reversed manner. Defaults to False.
            region_first (bool, optional): If the walk should go through the Statement first or the Region of a Statement first. Defaults to False.

        Yields:
            Iterator[Statement]: An iterator that yield Statements in the Block in the specified order.
        """
        for stmt in reversed(self.stmts) if reverse else self.stmts:
            yield from stmt.walk(reverse=reverse, region_first=region_first)

    def print_impl(self, printer: Printer) -> None:
        printer.plain_print(printer.state.block_id[self])
        printer.print_seq(
            [printer.state.ssa_id[arg] for arg in self.args],
            delim=", ",
            prefix="(",
            suffix="):",
            emit=printer.plain_print,
        )

        if printer.analysis is not None:
            with printer.indent(increase=4, mark=False):
                for arg in self.args:
                    printer.print_newline()
                    with printer.rich(style="warning"):
                        printer.print_analysis(
                            arg, prefix=f"{printer.state.ssa_id[arg]} --> "
                        )

        with printer.indent(increase=2, mark=False):
            for stmt in self.stmts:
                printer.print_newline()
                printer.print_stmt(stmt)

    def typecheck(self) -> None:
        """Checking the types of the Statments in the Block."""
        for stmt in self.stmts:
            stmt.typecheck()

    def verify(self) -> None:
        """Verify the correctness of the Block.

        Raises:
            VerificationError: If the Block is not correct.
        """
        from kirin.ir.nodes.stmt import Region

        if not isinstance(self.parent, Region):
            raise VerificationError(self, "Parent is not a region")

        for stmt in self.stmts:
            stmt.verify()



"""A function dialect that is compatible with python semantics."""

from kirin.dialects.func import (
    emit as emit,
    interp as interp,
    constprop as constprop,
    typeinfer as typeinfer,
)
from kirin.dialects.func.attrs import Signature as Signature, MethodType as MethodType
from kirin.dialects.func.stmts import (
    Call as Call,
    Invoke as Invoke,
    Lambda as Lambda,
    Return as Return,
    Function as Function,
    GetField as GetField,
    ConstantNone as ConstantNone,
)
from kirin.dialects.func.dialect import dialect as dialect



from typing import Any

from kirin.exceptions import VerificationError
from kirin.decl.emit.init import BaseModifier

from ._create_fn import create_fn
from ._set_new_attribute import set_new_attribute


class EmitVerify(BaseModifier):
    _VERIFICATION_ERROR = "_kirin_VerificationError"

    def emit_verify(self):
        verify_locals: dict[str, Any] = {
            self._VERIFICATION_ERROR: VerificationError,
        }
        body: list[str] = []
        for name in self.fields.blocks.keys():
            body.append(f"{self._self_name}.{name}.verify()")

        for name, f in self.fields.regions.items():
            body.append(f"{self._self_name}.{name}.verify()")
            if not f.multi:
                body.append(f"if len({self._self_name}.{name}.blocks) != 1:")
                body.append(
                    f"    raise {self._VERIFICATION_ERROR}({self._self_name},"
                    f" 'Invalid number of blocks for {name}')"
                )

        if (traits := getattr(self.cls, "traits", None)) is not None:
            for trait in traits:
                trait_obj = f"_kirin_verify_trait_{trait.__class__.__name__}"
                verify_locals.update({trait_obj: trait})
                body.append(f"{trait_obj}.verify({self._self_name})")

        # NOTE: we still need to generate this because it is abstract
        if not body:
            body.append("pass")

        set_new_attribute(
            self.cls,
            "verify",
            create_fn(
                name="_kirin_decl_verify",
                args=[self._self_name],
                body=body,
                globals=self.globals,
                locals=verify_locals,
                return_type=None,
            ),
        )



import ast
from dataclasses import dataclass


@dataclass
class SourceInfo:
    lineno: int
    col_offset: int
    end_lineno: int | None
    end_col_offset: int | None

    @classmethod
    def from_ast(cls, node: ast.AST, lineno_offset: int = 0, col_offset: int = 0):
        end_lineno = getattr(node, "end_lineno", None)
        end_col_offset = getattr(node, "end_col_offset", None)
        return cls(
            getattr(node, "lineno", 0) + lineno_offset,
            getattr(node, "col_offset", 0) + col_offset,
            end_lineno + lineno_offset if end_lineno is not None else None,
            end_col_offset + col_offset if end_col_offset is not None else None,
        )



from typing import Any

from kirin.ir import Block, Region
from kirin.ir.method import Method
from kirin.ir.nodes.stmt import Statement

from .base import BaseInterpreter
from .frame import Frame
from .value import Successor, YieldValue, ReturnValue, SpecialValue
from .exceptions import FuelExhaustedError


class Interpreter(BaseInterpreter[Frame[Any], Any]):
    """Concrete interpreter for the IR.

    This is a concrete interpreter for the IR. It evaluates the IR by
    executing the statements in the IR using a simple stack-based
    interpreter.
    """

    keys = ["main"]
    void = None

    def new_frame(self, code: Statement) -> Frame[Any]:
        return Frame.from_func_like(code)

    def run_method(
        self, method: Method, args: tuple[Any, ...]
    ) -> tuple[Frame[Any], Any]:
        return self.run_callable(method.code, (method,) + args)

    def run_ssacfg_region(
        self, frame: Frame[Any], region: Region
    ) -> tuple[Any, ...] | None | ReturnValue[Any]:
        block = region.blocks[0]
        while block is not None:
            results = self.run_block(frame, block)
            if isinstance(results, Successor):
                block = results.block
                frame.set_values(block.args, results.block_args)
            elif isinstance(results, ReturnValue):
                return results
            elif isinstance(results, YieldValue):
                return results.values
            else:
                return results
        return None  # region without terminator returns empty tuple

    def run_block(self, frame: Frame[Any], block: Block) -> SpecialValue[Any]:
        for stmt in block.stmts:
            if self.consume_fuel() == self.FuelResult.Stop:
                raise FuelExhaustedError("fuel exhausted")
            frame.stmt = stmt
            frame.lino = stmt.source.lineno if stmt.source else 0
            stmt_results = self.eval_stmt(frame, stmt)
            if isinstance(stmt_results, tuple):
                frame.set_values(stmt._results, stmt_results)
            elif stmt_results is None:
                continue  # empty result
            else:  # terminator
                return stmt_results
        return None



from kirin import types
from kirin.ir import (
    Pure,
    Method,
    Region,
    HasParent,
    MaybePure,
    Statement,
    ResultValue,
    ConstantLike,
    HasSignature,
    IsTerminator,
    SSACFGRegion,
    IsolatedFromAbove,
    SymbolOpInterface,
    CallableStmtInterface,
)
from kirin.decl import info, statement
from kirin.ir.ssa import SSAValue
from kirin.exceptions import VerificationError
from kirin.print.printer import Printer
from kirin.dialects.func.attrs import Signature, MethodType
from kirin.dialects.func.dialect import dialect

from .._pprint_helper import pprint_calllike


class FuncOpCallableInterface(CallableStmtInterface["Function"]):

    @classmethod
    def get_callable_region(cls, stmt: "Function") -> Region:
        return stmt.body


@statement(dialect=dialect)
class Function(Statement):
    name = "func"
    traits = frozenset(
        {
            IsolatedFromAbove(),
            SymbolOpInterface(),
            HasSignature(),
            FuncOpCallableInterface(),
            SSACFGRegion(),
        }
    )
    sym_name: str = info.attribute()
    """The symbol name of the function."""
    signature: Signature = info.attribute()
    body: Region = info.region(multi=True)

    def print_impl(self, printer: Printer) -> None:
        with printer.rich(style="keyword"):
            printer.print_name(self)
            printer.plain_print(" ")

        with printer.rich(style="symbol"):
            printer.plain_print(self.sym_name)

        printer.print_seq(self.signature.inputs, prefix="(", suffix=")", delim=", ")

        with printer.rich(style="comment"):
            printer.plain_print(" -> ")
            printer.print(self.signature.output)
            printer.plain_print(" ")

        printer.print(self.body)

        with printer.rich(style="comment"):
            printer.plain_print(f" // func.func {self.sym_name}")


@statement(dialect=dialect)
class Call(Statement):
    name = "call"
    traits = frozenset({MaybePure()})
    # not a fixed type here so just any
    callee: SSAValue = info.argument()
    inputs: tuple[SSAValue, ...] = info.argument()
    kwargs: tuple[str, ...] = info.attribute(default_factory=lambda: ())
    result: ResultValue = info.result()
    purity: bool = info.attribute(default=False)

    def print_impl(self, printer: Printer) -> None:
        pprint_calllike(self, printer.state.ssa_id[self.callee], printer)


@statement(dialect=dialect)
class ConstantNone(Statement):
    """A constant None value.

    This is mainly used to represent the None return value of a function
    to match Python semantics.
    """

    name = "const.none"
    traits = frozenset({Pure(), ConstantLike()})
    result: ResultValue = info.result(types.NoneType)


@statement(dialect=dialect, init=False)
class Return(Statement):
    name = "return"
    traits = frozenset({IsTerminator(), HasParent((Function,))})
    value: SSAValue = info.argument()

    def __init__(self, value_or_stmt: SSAValue | Statement | None = None) -> None:
        if isinstance(value_or_stmt, SSAValue):
            args = [value_or_stmt]
        elif isinstance(value_or_stmt, Statement):
            if len(value_or_stmt._results) == 1:
                args = [value_or_stmt._results[0]]
            else:
                raise ValueError(
                    f"expected a single result, got {len(value_or_stmt._results)} results from {value_or_stmt.name}"
                )
        elif value_or_stmt is None:
            args = []
        else:
            raise ValueError(f"expected SSAValue or Statement, got {value_or_stmt}")

        super().__init__(args=args, args_slice={"value": 0})

    def print_impl(self, printer: Printer) -> None:
        with printer.rich(style="keyword"):
            printer.print_name(self)

        if self.args:
            printer.plain_print(" ")
            printer.print_seq(self.args, delim=", ")

    def verify(self) -> None:
        if not self.args:
            raise VerificationError(
                self, "return statement must have at least one value"
            )

        if len(self.args) > 1:
            raise VerificationError(
                self,
                "return statement must have at most one value"
                ", wrap multiple values in a tuple",
            )


@statement(dialect=dialect)
class Lambda(Statement):
    name = "lambda"
    traits = frozenset(
        {
            Pure(),
            HasSignature(),
            SymbolOpInterface(),
            FuncOpCallableInterface(),
            SSACFGRegion(),
        }
    )
    sym_name: str = info.attribute()
    signature: Signature = info.attribute()
    captured: tuple[SSAValue, ...] = info.argument()
    body: Region = info.region(multi=True)
    result: ResultValue = info.result(MethodType)

    def verify(self) -> None:
        if self.body.blocks.isempty():
            raise VerificationError(self, "lambda body must not be empty")

    def print_impl(self, printer: Printer) -> None:
        with printer.rich(style="keyword"):
            printer.print_name(self)
        printer.plain_print(" ")

        with printer.rich(style="symbol"):
            printer.plain_print(self.sym_name)

        printer.print_seq(self.captured, prefix="(", suffix=")", delim=", ")

        with printer.rich(style="bright_black"):
            printer.plain_print(" -> ")
            printer.print(self.signature.output)

        printer.plain_print(" ")
        printer.print(self.body)

        with printer.rich(style="black"):
            printer.plain_print(f" // func.lambda {self.sym_name}")


@statement(dialect=dialect)
class GetField(Statement):
    name = "getfield"
    traits = frozenset({Pure()})
    obj: SSAValue = info.argument(MethodType)
    field: int = info.attribute()
    # NOTE: mypy somehow doesn't understand default init=False
    result: ResultValue = info.result(init=False)

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self)
        printer.plain_print(
            "(", printer.state.ssa_id[self.obj], ", ", str(self.field), ")"
        )
        with printer.rich(style="black"):
            printer.plain_print(" : ")
            printer.print(self.result.type)


@statement(dialect=dialect)
class Invoke(Statement):
    name = "invoke"
    traits = frozenset({MaybePure()})
    callee: Method = info.attribute()
    inputs: tuple[SSAValue, ...] = info.argument()
    kwargs: tuple[str, ...] = info.attribute()
    result: ResultValue = info.result()
    purity: bool = info.attribute(default=False)

    def print_impl(self, printer: Printer) -> None:
        pprint_calllike(self, self.callee.sym_name, printer)

    def verify(self) -> None:
        if self.kwargs:
            for name in self.kwargs:
                if name not in self.callee.arg_names:
                    raise VerificationError(
                        self,
                        f"method {self.callee.sym_name} does not have argument {name}",
                    )
        elif len(self.callee.arg_names) - 1 != len(self.args):
            raise VerificationError(
                self,
                f"expected {len(self.callee.arg_names)} arguments, got {len(self.args)}",
            )



"""Assignment dialect for Python.

This module contains the dialect for the Python assignment statement, including:

- Statements: `Alias`, `SetItem`.
- The lowering pass for the assignments.
- The concrete implementation of the assignment statements.

This dialects maps Python assignment syntax.
"""

import ast

from kirin import ir, types, interp, lowering, exceptions
from kirin.decl import info, statement
from kirin.print import Printer

dialect = ir.Dialect("py.assign")

T = types.TypeVar("T")


@statement(dialect=dialect)
class Alias(ir.Statement):
    name = "alias"
    traits = frozenset({ir.Pure(), ir.FromPythonCall()})
    value: ir.SSAValue = info.argument(T)
    target: ir.PyAttr[str] = info.attribute()
    result: ir.ResultValue = info.result(T)

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self)
        printer.plain_print(" ")
        with printer.rich(style="symbol"):
            printer.plain_print(self.target.data)

        with printer.rich(style="keyword"):
            printer.plain_print(" = ")

        printer.print(self.value)


@statement(dialect=dialect)
class SetItem(ir.Statement):
    name = "setitem"
    traits = frozenset({ir.FromPythonCall()})
    obj: ir.SSAValue = info.argument(print=False)
    value: ir.SSAValue = info.argument(print=False)
    index: ir.SSAValue = info.argument(print=False)


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(Alias)
    def alias(self, interp, frame: interp.Frame, stmt: Alias):
        return (frame.get(stmt.value),)

    @interp.impl(SetItem)
    def setindex(self, interp, frame: interp.Frame, stmt: SetItem):
        frame.get(stmt.obj)[frame.get(stmt.index)] = frame.get(stmt.value)
        return (None,)


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_Assign(
        self, state: lowering.LoweringState, node: ast.Assign
    ) -> lowering.Result:
        results: lowering.Result = state.visit(node.value)
        assert len(node.targets) == len(
            results
        ), "number of targets and results do not match"

        current_frame = state.current_frame
        match node:
            case ast.Assign(
                targets=[ast.Name(lhs_name, ast.Store())], value=ast.Name(_, ast.Load())
            ):
                stmt = Alias(
                    value=results[0], target=ir.PyAttr(lhs_name)
                )  # NOTE: this is guaranteed to be one result
                stmt.result.name = lhs_name
                current_frame.defs[lhs_name] = state.append_stmt(stmt).result
            case _:
                for target, value in zip(node.targets, results.values):
                    match target:
                        # NOTE: if the name exists new ssa value will be
                        # used in the future to shadow the old one
                        case ast.Name(name, ast.Store()):
                            value.name = name
                            current_frame.defs[name] = value
                        case ast.Subscript(obj, slice):
                            obj = state.visit(obj).expect_one()
                            slice = state.visit(slice).expect_one()
                            stmt = SetItem(obj=obj, index=slice, value=value)
                            state.append_stmt(stmt)
                        case _:
                            raise exceptions.DialectLoweringError(
                                f"unsupported target {target}"
                            )
        return lowering.Result()  # python assign does not have value



"""Lowering Python AST to cf dialect."""

import ast

from kirin import ir, types
from kirin.dialects import cf, py
from kirin.lowering import Frame, Result, FromPythonAST, LoweringState
from kirin.exceptions import DialectLoweringError

dialect = ir.Dialect("lowering.cf")


@dialect.register
class CfLowering(FromPythonAST):

    def lower_Pass(self, state: LoweringState, node: ast.Pass) -> Result:
        state.append_stmt(
            cf.Branch(arguments=(), successor=state.current_frame.next_block)
        )
        return Result()

    def lower_For(self, state: LoweringState, node: ast.For) -> Result:
        yields: list[str] = []

        def new_block_arg_if_inside_loop(frame: Frame, capture: ir.SSAValue):
            if not capture.name:
                raise DialectLoweringError("unexpected loop variable captured")
            yields.append(capture.name)
            return frame.entr_block.args.append_from(capture.type, capture.name)

        frame = state.current_frame
        iterable = state.visit(node.iter).expect_one()
        iter_stmt = frame.append_stmt(py.iterable.Iter(iterable))
        none_stmt = frame.append_stmt(py.Constant(None))

        body_frame = state.push_frame(
            Frame.from_stmts(
                node.body,
                state,
                region=state.current_frame.curr_region,
                globals=state.current_frame.globals,
                capture_callback=new_block_arg_if_inside_loop,
            )
        )
        next_value = body_frame.entr_block.args.append_from(types.Any, "next_value")
        py.unpack.unpacking(state, node.target, next_value)
        state.exhaust(body_frame)
        self.branch_next_if_not_terminated(body_frame)
        yield_args = tuple(body_frame.get_scope(name) for name in yields)
        next_stmt = py.iterable.Next(iter_stmt.iter)
        cond_stmt = py.cmp.Is(next_stmt.value, none_stmt.result)
        body_frame.next_block.stmts.append(next_stmt)
        body_frame.next_block.stmts.append(cond_stmt)
        body_frame.next_block.stmts.append(
            cf.ConditionalBranch(
                cond_stmt.result,
                yield_args,
                (next_stmt.value,) + yield_args,
                then_successor=frame.next_block,
                else_successor=body_frame.entr_block,
            )
        )
        state.pop_frame()

        # insert the branch to the entrance of the loop (the code block before loop)
        next_stmt = frame.append_stmt(py.iterable.Next(iter_stmt.iter))
        cond_stmt = frame.append_stmt(py.cmp.Is(next_stmt.value, none_stmt.result))
        yield_args = tuple(frame.get_scope(name) for name in yields)
        frame.append_stmt(
            cf.ConditionalBranch(
                cond_stmt.result,
                yield_args,
                (next_stmt.value,) + yield_args,
                then_successor=frame.next_block,  # empty iterator
                else_successor=body_frame.entr_block,
            )
        )

        frame.jump_next()
        for name, arg in zip(yields, yield_args):
            input = frame.curr_block.args.append_from(arg.type, name)
            frame.defs[name] = input
        return Result()

    def lower_If(self, state: LoweringState, node: ast.If) -> Result:
        cond = state.visit(node.test).expect_one()
        frame = state.current_frame
        before_block = frame.curr_block
        if_frame = state.push_frame(
            Frame.from_stmts(
                node.body,
                state,
                region=frame.curr_region,
                globals=frame.globals,
            )
        )
        true_cond = if_frame.entr_block.args.append_from(types.Bool, cond.name)
        if cond.name:
            if_frame.defs[cond.name] = true_cond
        state.exhaust()
        self.branch_next_if_not_terminated(if_frame)
        state.pop_frame()

        else_frame = state.push_frame(
            Frame.from_stmts(
                node.orelse,
                state,
                region=frame.curr_region,
                globals=frame.globals,
            )
        )
        true_cond = else_frame.entr_block.args.append_from(types.Bool, cond.name)
        if cond.name:
            else_frame.defs[cond.name] = true_cond
        state.exhaust()
        self.branch_next_if_not_terminated(else_frame)
        state.pop_frame()

        after_frame = state.push_frame(
            Frame.from_stmts(
                frame.stream.split(),
                state,
                region=frame.curr_region,
                globals=frame.globals,
            )
        )

        after_frame.defs.update(frame.defs)
        phi: set[str] = set()
        for name in if_frame.defs.keys():
            if frame.get(name):
                phi.add(name)
            elif name in else_frame.defs:
                phi.add(name)

        for name in else_frame.defs.keys():
            if frame.get(name):  # not defined in if_frame
                phi.add(name)

        for name in phi:
            after_frame.defs[name] = after_frame.entr_block.args.append_from(
                types.Any, name
            )

        state.exhaust()
        self.branch_next_if_not_terminated(after_frame)
        after_frame.next_block.stmts.append(
            cf.Branch(arguments=(), successor=frame.next_block)
        )
        state.pop_frame()

        if_args = []
        for name in phi:
            if value := if_frame.get(name):
                if_args.append(value)
            else:
                raise DialectLoweringError(f"undefined variable {name} in if branch")

        else_args = []
        for name in phi:
            if value := else_frame.get(name):
                else_args.append(value)
            else:
                raise DialectLoweringError(f"undefined variable {name} in else branch")

        if_frame.next_block.stmts.append(
            cf.Branch(
                arguments=tuple(if_args),
                successor=after_frame.entr_block,
            )
        )
        else_frame.next_block.stmts.append(
            cf.Branch(
                arguments=tuple(else_args),
                successor=after_frame.entr_block,
            )
        )
        before_block.stmts.append(
            cf.ConditionalBranch(
                cond=cond,
                then_arguments=(cond,),
                then_successor=if_frame.entr_block,
                else_arguments=(cond,),
                else_successor=else_frame.entr_block,
            )
        )
        frame.jump_next()
        return Result()

    def branch_next_if_not_terminated(self, frame: Frame):
        """Branch to the next block if the current block is not terminated.

        This must be used after exhausting the current frame and before popping the frame.
        """
        if not frame.curr_block.last_stmt or not frame.curr_block.last_stmt.has_trait(
            ir.IsTerminator
        ):
            frame.curr_block.stmts.append(
                cf.Branch(arguments=(), successor=frame.next_block)
            )

    def current_block_terminated(self, frame: Frame):
        return frame.curr_block.last_stmt and frame.curr_block.last_stmt.has_trait(
            ir.IsTerminator
        )



import pytest

from kirin.types import (
    Int,
    Bool,
    Dict,
    Float,
    Slice,
    Tuple,
    Union,
    String,
    Vararg,
    AnyType,
    Literal,
    PyClass,
    TypeVar,
    NoneType,
    BottomType,
    TypeAttribute,
)


class Base:
    pass


class Derived(Base):
    pass


def test_union():
    assert PyClass(int) | PyClass(float) == Union(PyClass(int), PyClass(float))
    assert Union(PyClass(int), PyClass(int)) == PyClass(int)
    assert Union(PyClass(int), PyClass(float)) == Union(PyClass(int), PyClass(float))
    assert Union(Int, Float, BottomType()).is_equal(Union(Int, Float))
    assert hash(Union(PyClass(int), PyClass(float))) == hash(
        Union(PyClass(int), PyClass(float))
    )
    assert Union(PyClass(int), PyClass(float)) == Union(PyClass(float), PyClass(int))
    assert hash(Union(PyClass(int), PyClass(float))) == hash(
        Union(PyClass(float), PyClass(int))
    )
    assert Union(Union(Int, Float), BottomType()) == Union(Int, Float)
    assert Union(PyClass(int), AnyType()) == AnyType()
    assert Union(AnyType(), PyClass(int)) == AnyType()
    assert Union(BottomType(), PyClass(int)) == PyClass(int)
    assert Union(PyClass(int), BottomType()) == PyClass(int)
    assert PyClass(Derived).is_subseteq(PyClass(Base))
    assert Union(PyClass(Derived), PyClass(Base)) == PyClass(Base)
    assert AnyType() is AnyType()
    assert BottomType() is BottomType()
    t = Int.join(Float).join(String)
    assert t.is_subseteq(Int.join(Float).join(String))


def test_meet():
    assert PyClass(int).meet(PyClass(float)) == BottomType()
    assert PyClass(int).meet(PyClass(int)) == PyClass(int)
    assert PyClass(int).meet(AnyType()) == PyClass(int)
    assert AnyType().meet(PyClass(int)) == PyClass(int)
    assert BottomType().meet(PyClass(int)) == BottomType()
    assert PyClass(Base).meet(PyClass(Derived)) == PyClass(Derived)


def test_literal():
    assert Literal(Int) == Int
    assert Literal("aaa").join(Literal("bbb")) == Union(Literal("bbb"), Literal("aaa"))
    assert Literal("aaa").meet(Literal("bbb")) == BottomType()
    assert Literal("aaa").meet(Literal("aaa")) == Literal("aaa")
    assert Literal("aaa").is_subseteq(Literal("aaa") | String)
    assert Int.is_subseteq(Literal("aaa")) is False
    assert Tuple[Int].is_subseteq(Literal("aaa")) is False


def test_singleton():
    assert hash(AnyType()) == hash(AnyType())
    assert hash(AnyType()) == id(AnyType())
    assert hash(BottomType()) == hash(BottomType())
    assert hash(BottomType()) == id(BottomType())
    assert NoneType is NoneType
    assert Int is PyClass(int)
    assert Float is PyClass(float)
    assert String is PyClass(str)
    assert Bool is PyClass(bool)
    assert Literal("aaa") is Literal("aaa")


def test_generic_is_subseteq():
    assert Tuple[Literal("aaa")].is_subseteq(Tuple[Literal("aaa")])
    assert Tuple[Vararg(Int)][Int, Int] == Tuple[Int, Int]
    assert hash(Tuple[Int, Int]) == hash(Tuple[Int, Int])
    assert Tuple[Int, Vararg(Int)][Int, Int] == Tuple[Int, Int]
    assert Tuple[Int, Int].is_subseteq(Tuple[TypeVar("T"), Int])
    assert Dict[Int, Int].is_subseteq(Dict[TypeVar("K"), TypeVar("V")])
    assert Dict[Int, Int].is_subseteq(Dict)
    assert Dict[Int, Int].is_subseteq(Dict[Int])
    assert not Dict[Int, Int].is_subseteq(Dict[Float])
    assert PyClass(slice).is_subseteq(Slice)
    assert TypeVar("T", Int).is_subseteq(Int | String)

    with pytest.raises(TypeError):
        Tuple[Vararg(Int)][Int, Float]

    with pytest.raises(TypeError):
        Tuple[Vararg(Int), Int]


def test_generic_topbottom():
    t = Union(Int, Float)
    assert t.join(TypeAttribute.bottom()).is_subseteq(t)
    assert t.meet(TypeAttribute.bottom()).is_subseteq(TypeAttribute.bottom())
    assert t.join(TypeAttribute.top()).is_equal(TypeAttribute.top())
    assert t.meet(TypeAttribute.top()).is_equal(t)



from kirin import ir
from kirin.decl import info, statement
from kirin.prelude import python_no_opt
from kirin.dialects import cf, py, func
from kirin.lowering import Lowering

dialect = ir.Dialect("test")


@statement(dialect=dialect)
class Adjoint(ir.Statement):
    traits = frozenset({ir.FromPythonWithSingleItem()})
    body: ir.Region = info.region()
    result: ir.ResultValue = info.result()


def with_example(x):
    y = 1
    with Adjoint() as f:  # type: ignore
        y = x + 1
    return y, f


def test_with_lowering():
    lowering = Lowering(python_no_opt.union([cf, func, dialect]))
    code = lowering.run(with_example)
    code.print()
    assert isinstance(code, func.Function)
    stmts = code.body.blocks[0].stmts
    assert isinstance(stmts.at(0), py.Constant)
    adjoint = stmts.at(1)
    assert isinstance(adjoint, Adjoint)
    assert len(adjoint.body.blocks) == 1
    add = adjoint.body.blocks[0].stmts.at(1)
    assert isinstance(add, py.Add)
    assert isinstance(add.lhs, ir.BlockArgument)
    assert isinstance(add.rhs, ir.SSAValue)
    assert adjoint.result.name == "f"



from typing import Any

from kirin import ir
from kirin.decl import info, fields, statement
from kirin.dialects.py import types

T = types.PyTypeVar("T")


@statement(repr=True)
class TestStatement(ir.Statement):
    name = "constant"
    traits = frozenset({ir.Pure(), ir.ConstantLike()})

    # args
    noinfo: ir.SSAValue
    vararg_noinfo: tuple[ir.SSAValue, ...]
    xxx: ir.SSAValue = info.argument(T)
    xxx_Any: ir.SSAValue = info.argument()
    xxx_vararg: tuple[ir.SSAValue, ...] = info.argument()

    # results
    xxx_result: ir.ResultValue = info.result(T)

    # attributes
    xxx_property: Any = info.attribute(T, default="", property=True)
    xxx_attribute: Any = info.attribute(T)
    xxx_dict: dict[str, int] = info.attribute()

    # regions
    xxx_region_noinfo: ir.Region = info.region()
    xxx_region: ir.Region = info.region(default_factory=ir.Region)
    xxx_region_multi: ir.Region = info.region(default_factory=ir.Region, multi=True)

    # blocks
    block_noinfo: ir.Block = info.block()
    block_default: ir.Block = info.block(default_factory=ir.Block)


fields(TestStatement)

print(TestStatement._arg_groups)
stmt = TestStatement(
    ir.TestValue(),
    (ir.TestValue(), ir.TestValue()),
    ir.TestValue(),
    xxx_Any=ir.TestValue(),
    xxx_vararg=(ir.TestValue(), ir.TestValue()),
    xxx_attribute=2,
    xxx_property=1,
    xxx_dict={"a": 1},
    xxx_region_noinfo=ir.Region(),
    block_noinfo=ir.Block(),
)

print(stmt)
# print(stmt.xxx_property)
# print(stmt.properties)
# print(stmt.xxx_Any)
# print(stmt.xxx_vararg)
# print(stmt.block_noinfo)
# print(stmt.xxx_region_multi)
# print(stmt.xxx_property)
# stmt.print()
# print(stmt.__repr__())

# ff = fields(TestStatement)
# from rich import print

# print(ff.attributes["xxx_dict"])
# # print(ff.args)
# # print(ff.results)
# # print(ff.attributes)
# # print(ff.regions)



from kirin import ir, types
from kirin.decl import info, statement
from kirin.prelude import basic_no_opt
from kirin.analysis import const
from kirin.dialects import ilist


class TestLattice:

    def test_meet(self):
        assert const.Unknown().meet(const.Unknown()) == const.Unknown()
        assert const.Unknown().meet(const.Bottom()) == const.Bottom()
        assert const.Unknown().meet(const.Value(1)) == const.Value(1)
        assert const.Unknown().meet(
            const.PartialTuple((const.Value(1), const.Bottom()))
        ) == const.PartialTuple((const.Value(1), const.Bottom()))
        assert const.Bottom().meet(const.Unknown()) == const.Bottom()
        assert const.Bottom().meet(const.Bottom()) == const.Bottom()
        assert const.Bottom().meet(const.Value(1)) == const.Bottom()
        assert (
            const.Bottom().meet(const.PartialTuple((const.Value(1), const.Bottom())))
            == const.Bottom()
        )
        assert const.Value(1).meet(const.Unknown()) == const.Value(1)
        assert const.Value(1).meet(const.Bottom()) == const.Bottom()
        assert const.Value(1).meet(const.Value(1)) == const.Value(1)
        assert (
            const.Value(1).meet(const.PartialTuple((const.Value(1), const.Bottom())))
            == const.Bottom()
        )
        assert const.PartialTuple((const.Value(1), const.Bottom())).meet(
            const.Unknown()
        ) == const.PartialTuple((const.Value(1), const.Bottom()))
        assert (
            const.PartialTuple((const.Value(1), const.Bottom())).meet(const.Bottom())
            == const.Bottom()
        )
        assert (
            const.PartialTuple((const.Value(1), const.Bottom())).meet(const.Value(1))
            == const.Bottom()
        )
        assert const.PartialTuple((const.Value(1), const.Bottom())).meet(
            const.Value((1, 2))
        ) == const.PartialTuple((const.Value(1), const.Bottom()))
        assert const.PartialTuple((const.Value(1), const.Bottom())).meet(
            const.PartialTuple((const.Value(1), const.Bottom()))
        ) == const.PartialTuple((const.Value(1), const.Bottom()))

    def test_join(self):
        assert const.Unknown().join(const.Unknown()) == const.Unknown()
        assert const.Unknown().join(const.Bottom()) == const.Unknown()
        assert const.Unknown().join(const.Value(1)) == const.Unknown()
        assert (
            const.Unknown().join(const.PartialTuple((const.Value(1), const.Bottom())))
            == const.Unknown()
        )
        assert const.Bottom().join(const.Unknown()) == const.Unknown()
        assert const.Bottom().join(const.Bottom()) == const.Bottom()
        assert const.Bottom().join(const.Value(1)) == const.Value(1)
        assert const.Bottom().join(
            const.PartialTuple((const.Value(1), const.Bottom()))
        ) == const.PartialTuple((const.Value(1), const.Bottom()))
        assert const.PartialTuple((const.Value(1), const.Bottom())).join(
            const.Value((1, 2))
        ) == const.PartialTuple((const.Value(1), const.Value(2)))
        assert const.Value(1).join(const.Unknown()) == const.Unknown()
        assert const.Value(1).join(const.Bottom()) == const.Value(1)
        assert const.Value(1).join(const.Value(1)) == const.Value(1)
        assert const.Value(1).join(const.Value(2)) == const.Unknown()
        assert (
            const.Value(1).join(const.PartialTuple((const.Value(1), const.Bottom())))
            == const.Unknown()
        )

    def test_is_equal(self):
        assert const.Unknown().is_equal(const.Unknown())
        assert not const.Unknown().is_equal(const.Bottom())
        assert not const.Unknown().is_equal(const.Value(1))
        assert const.Bottom().is_equal(const.Bottom())
        assert not const.Bottom().is_equal(const.Value(1))
        assert const.Value(1).is_equal(const.Value(1))
        assert not const.Value(1).is_equal(const.Value(2))
        assert const.PartialTuple((const.Value(1), const.Bottom())).is_equal(
            const.PartialTuple((const.Value(1), const.Bottom()))
        )
        assert not const.PartialTuple((const.Value(1), const.Bottom())).is_equal(
            const.PartialTuple((const.Value(1), const.Value(2)))
        )

    def test_partial_tuple(self):
        pt1 = const.PartialTuple((const.Value(1), const.Bottom()))
        pt2 = const.PartialTuple((const.Value(1), const.Bottom()))
        assert pt1.is_equal(pt2)
        assert pt1.is_subseteq(pt2)
        assert pt1.join(pt2) == pt1
        assert pt1.meet(pt2) == pt1
        pt2 = const.PartialTuple((const.Value(1), const.Value(2)))
        assert not pt1.is_equal(pt2)
        assert pt1.is_subseteq(pt2)
        assert pt1.join(pt2) == const.PartialTuple((const.Value(1), const.Value(2)))
        assert pt1.meet(pt2) == const.PartialTuple((const.Value(1), const.Bottom()))
        pt2 = const.PartialTuple((const.Value(1), const.Bottom()))
        assert pt1.is_equal(pt2)
        assert pt1.is_subseteq(pt2)
        assert pt1.join(pt2) == pt1
        assert pt1.meet(pt2) == pt1
        pt2 = const.PartialTuple((const.Value(1), const.Unknown()))
        assert not pt1.is_equal(pt2)
        assert pt1.is_subseteq(pt2)
        assert pt1.join(pt2) == pt2
        assert pt1.meet(pt2) == pt1


@basic_no_opt
def foo(x):
    return x + 1


@basic_no_opt
def goo(x):
    return foo(2), foo(x)


@basic_no_opt
def main():
    return goo(3)


@basic_no_opt
def bar(x):
    return goo(x)[0]


@basic_no_opt
def ntuple(len: int):
    if len == 0:
        return ()
    return (0,) + ntuple(len - 1)


@basic_no_opt
def recurse():
    return ntuple(3)


def test_constprop():
    infer = const.Propagate(basic_no_opt)
    frame, ret = infer.run_analysis(main)
    assert ret == const.Value((3, 4))
    assert len(frame.entries) == 3

    frame, ret = infer.run_analysis(goo)
    assert ret == const.PartialTuple((const.Value(3), const.Unknown()))
    assert len(frame.entries) == 6
    block = goo.callable_region.blocks[0]
    assert frame.entries[block.stmts.at(1).results[0]] == const.Value(3)
    assert frame.entries[block.stmts.at(2).results[0]] == const.Unknown()
    assert frame.entries[block.stmts.at(3).results[0]] == const.PartialTuple(
        (const.Value(3), const.Unknown())
    )

    _, ret = infer.run_analysis(bar)
    assert ret == const.Value(3)

    _, ret = infer.run_analysis(foo)
    assert ret == const.Unknown()
    _, ret = infer.run_analysis(recurse)
    assert ret == const.Value((0, 0, 0))


@basic_no_opt
def myfunc(x1: int) -> int:
    return x1 * 2


@basic_no_opt
def _for_loop_test_constp(
    cntr: int,
    x: tuple,
    n_range: int,
):
    if cntr < n_range:
        pos = myfunc(cntr)
        x = x + (cntr, pos)
        return _for_loop_test_constp(
            cntr=cntr + 1,
            x=x,
            n_range=n_range,
        )
    else:
        return x


def test_issue_40():
    constprop = const.Propagate(basic_no_opt)
    frame, ret = constprop.run_analysis(
        _for_loop_test_constp,
        (
            const.Value(0),
            const.Value(()),
            const.Value(5),
        ),
    )
    assert isinstance(ret, const.Value)
    assert ret.data == _for_loop_test_constp(cntr=0, x=(), n_range=5)


dummy_dialect = ir.Dialect("dummy")


@statement(dialect=dummy_dialect)
class DummyStatement(ir.Statement):
    name = "dummy"
    traits = frozenset({ir.FromPythonCall()})


def test_intraprocedure_side_effect():

    @basic_no_opt.add(dummy_dialect)
    def side_effect_return_none():
        DummyStatement()

    @basic_no_opt.add(dummy_dialect)
    def side_effect_intraprocedure(cond: bool):
        if cond:
            return side_effect_return_none()
        else:
            x = (1, 2, 3)
            return x

    constprop = const.Propagate(basic_no_opt.add(dummy_dialect))
    frame, ret = constprop.run_analysis(side_effect_intraprocedure)
    new_tuple = (
        side_effect_intraprocedure.callable_region.blocks[2].stmts.at(3).results[0]
    )
    assert isinstance(ret, const.Unknown)
    assert frame.entries[new_tuple] == const.Value((1, 2, 3))


def test_interprocedure_true_branch():
    @basic_no_opt.add(dummy_dialect)
    def side_effect_maybe_return_none(cond: bool):
        if cond:
            return
        else:
            DummyStatement()
            return

    @basic_no_opt.add(dummy_dialect)
    def side_effect_true_branch_const(cond: bool):
        if cond:
            return side_effect_maybe_return_none(cond)
        else:
            return cond

    constprop = const.Propagate(basic_no_opt.add(dummy_dialect))
    frame, ret = constprop.run_analysis(side_effect_true_branch_const)
    assert isinstance(ret, const.Unknown)  # instead of NotPure
    true_branch = side_effect_true_branch_const.callable_region.blocks[1]
    assert frame.entries[true_branch.stmts.at(0).results[0]] == const.Value(None)


def test_non_pure_recursion():
    @basic_no_opt
    def for_loop_append(cntr: int, x: ilist.IList, n_range: int):
        if cntr < n_range:
            for_loop_append(cntr + 1, x + [cntr], n_range)

        return x

    constprop = const.Propagate(basic_no_opt)
    frame, _ = constprop.run_analysis(for_loop_append)
    stmt = for_loop_append.callable_region.blocks[1].stmts.at(3)
    assert isinstance(frame.entries[stmt.results[0]], const.Unknown)


def test_closure_prop():
    dialect = ir.Dialect("dummy2")

    @statement(dialect=dialect)
    class DummyStmt2(ir.Statement):
        name = "dummy2"
        traits = frozenset({ir.FromPythonCall()})
        value: ir.SSAValue = info.argument(types.Int)
        result: ir.ResultValue = info.result(types.Int)

    @basic_no_opt.add(dialect)
    def non_const_closure(x: int, y: int):
        def inner():
            if False:
                return x + y
            else:
                return 2

        return inner

    @basic_no_opt.add(dialect)
    def non_pure(x: int, y: int):
        def inner():
            if False:
                return x + y
            else:
                DummyStmt2(1)  # type: ignore
                return 2

        return inner

    @basic_no_opt.add(dialect)
    def main():
        x = DummyStmt2(1)  # type: ignore
        x = non_const_closure(x, x)  # type: ignore
        return x()

    constprop = const.Propagate(basic_no_opt.add(dialect))
    frame, ret = constprop.run_analysis(main)
    main.print(analysis=frame.entries)
    stmt = main.callable_region.blocks[0].stmts.at(3)
    call_result = frame.entries[stmt.results[0]]
    assert isinstance(call_result, const.Value)
    assert call_result.data == 2

    @basic_no_opt.add(dialect)
    def main2():
        x = DummyStmt2(1)  # type: ignore
        x = non_pure(x, x)  # type: ignore
        return x()

    constprop = const.Propagate(basic_no_opt.add(dialect))
    frame, _ = constprop.run_analysis(main2)
    main2.print(analysis=frame.entries)
    stmt = main2.callable_region.blocks[0].stmts.at(3)
    call_result = frame.entries[stmt.results[0]]
    assert isinstance(call_result, const.Value)


def test_issue_300():
    @basic_no_opt
    def my_ps(val: float):
        def my_ps_impl():
            return val * 0.3

        return my_ps_impl

    @basic_no_opt
    def my_ps2(val: float):
        my_ps_impl = my_ps(val)
        return my_ps_impl()

    prop = const.Propagate(basic_no_opt)
    frame, ret = prop.run_analysis(my_ps2)
    invoke = my_ps2.callable_region.blocks[0].stmts.at(0)
    call = my_ps2.callable_region.blocks[0].stmts.at(1)
    assert invoke in frame.should_be_pure
    assert call in frame.should_be_pure



import pytest

from kirin import ir, types
from kirin.prelude import python_no_opt
from kirin.dialects import cf, func
from kirin.lowering import Lowering
from kirin.exceptions import DialectLoweringError

lowering = Lowering(python_no_opt)


def test_basic_func():
    def single(n):
        return n + 1

    code = lowering.run(single)
    assert isinstance(code, func.Function)
    assert len(code.body.blocks) == 1
    assert isinstance(code.body.blocks[0].last_stmt, func.Return)

    def single_2(n):
        return n + 1, n + 2

    code = lowering.run(single_2)
    assert isinstance(code, func.Function)
    assert len(code.body.blocks) == 1
    assert isinstance(code.body.blocks[0].last_stmt, func.Return)
    assert code.body.blocks[0].last_stmt.args[0].type.is_subseteq(types.Tuple)


def test_recursive_func():
    def recursive(n):
        if n == 0:
            return 0
        return recursive(n - 1)

    code = lowering.run(recursive)
    assert isinstance(code, func.Function)
    assert len(code.body.blocks) == 3
    assert isinstance(code.body.blocks[0].last_stmt, cf.ConditionalBranch)
    assert isinstance(code.body.blocks[2].stmts.at(2), func.Call)
    stmt: func.Call = code.body.blocks[2].stmts.at(2)  # type: ignore
    assert isinstance(stmt.callee, ir.BlockArgument)
    assert stmt.callee.type.is_subseteq(func.MethodType)


def test_invalid_func_call():

    def undefined(n):
        return foo(n - 1)  # type: ignore # noqa: F821

    with pytest.raises(DialectLoweringError):
        lowering.run(undefined)

    def calling_python(n):
        return print(n)

    with pytest.raises(
        DialectLoweringError,
        match="`lower_Call_print` is not implemented for builtin function `print`.",
    ):
        lowering.run(calling_python)


def test_func_call():
    def callee(n):
        return n + 1

    def caller(n):
        return callee(n)

    code = lowering.run(callee)
    callee = ir.Method(None, callee, "callee", ["n"], lowering.dialects, code)
    code = lowering.run(caller, globals={"callee": callee})
    assert isinstance(code, func.Function)
    assert len(code.body.blocks) == 1
    stmt = code.body.blocks[0].stmts.at(0)
    assert isinstance(stmt, func.Invoke)
    assert isinstance(stmt.callee, ir.Method)


def test_func_kw_call():
    def callee(n, m):
        return n + m

    def caller(n, m):  # type: ignore
        return callee(n=n, m=m)

    code = lowering.run(callee)
    callee = ir.Method(None, callee, "callee", ["n", "m"], lowering.dialects, code)
    code = lowering.run(caller, globals={"callee": callee})
    assert isinstance(code, func.Function)
    assert len(code.body.blocks) == 1
    stmt = code.body.blocks[0].stmts.at(0)
    assert isinstance(stmt, func.Invoke)
    assert stmt.kwargs == ("n", "m")

    def caller(n, m):
        return callee(n, m=m)

    code = lowering.run(caller, globals={"callee": callee})
    assert isinstance(code, func.Function)
    assert len(code.body.blocks) == 1
    stmt = code.body.blocks[0].stmts.at(0)
    assert isinstance(stmt, func.Invoke)
    assert stmt.kwargs == ("m",)



from kirin import interp
from kirin.emit.julia import EmitJulia, EmitStrFrame

from . import stmts
from ._dialect import dialect


@dialect.register(key="emit.julia")
class JuliaTable(interp.MethodTable):

    @interp.impl(stmts.Eq)
    def emit_Eq(self, emit: EmitJulia, frame: EmitStrFrame, stmt: stmts.Eq):
        return emit.emit_binaryop(frame, "==", stmt.lhs, stmt.rhs, stmt.result)

    @interp.impl(stmts.GtE)
    def emit_GtE(self, emit: EmitJulia, frame: EmitStrFrame, stmt: stmts.GtE):
        return emit.emit_binaryop(frame, ">=", stmt.lhs, stmt.rhs, stmt.result)

    @interp.impl(stmts.LtE)
    def emit_LtE(self, emit: EmitJulia, frame: EmitStrFrame, stmt: stmts.LtE):
        return emit.emit_binaryop(frame, "<=", stmt.lhs, stmt.rhs, stmt.result)

    @interp.impl(stmts.NotEq)
    def emit_NotEq(self, emit: EmitJulia, frame: EmitStrFrame, stmt: stmts.NotEq):
        return emit.emit_binaryop(frame, "!=", stmt.lhs, stmt.rhs, stmt.result)

    @interp.impl(stmts.Gt)
    def emit_Gt(self, emit: EmitJulia, frame: EmitStrFrame, stmt: stmts.Gt):
        return emit.emit_binaryop(frame, ">", stmt.lhs, stmt.rhs, stmt.result)

    @interp.impl(stmts.Lt)
    def emit_Lt(self, emit: EmitJulia, frame: EmitStrFrame, stmt: stmts.Lt):
        return emit.emit_binaryop(frame, "<", stmt.lhs, stmt.rhs, stmt.result)



"""The unpack dialect for Python.

This module contains the dialect for the Python unpack semantics, including:

- The `Unpack` statement class.
- The lowering pass for the unpack statement.
- The concrete implementation of the unpack statement.
- The type inference implementation of the unpack statement.
- A helper function `unpacking` for unpacking Python AST nodes during lowering.
"""

import ast

from kirin import ir, types, interp, lowering
from kirin.decl import info, statement
from kirin.print import Printer
from kirin.exceptions import DialectLoweringError

dialect = ir.Dialect("py.unpack")


@statement(dialect=dialect, init=False)
class Unpack(ir.Statement):
    value: ir.SSAValue = info.argument(types.Any)
    names: tuple[str | None, ...] = info.attribute()

    def __init__(self, value: ir.SSAValue, names: tuple[str | None, ...]):
        result_types = [types.Any] * len(names)
        super().__init__(
            args=(value,),
            result_types=result_types,
            args_slice={"value": 0},
            attributes={"names": ir.PyAttr(names)},
        )
        for result, name in zip(self.results, names):
            result.name = name

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self)
        printer.plain_print(" ")
        printer.print(self.value)


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(Unpack)
    def unpack(self, interp: interp.Interpreter, frame: interp.Frame, stmt: Unpack):
        return tuple(frame.get(stmt.value))


@dialect.register(key="typeinfer")
class TypeInfer(interp.MethodTable):

    @interp.impl(Unpack)
    def unpack(self, interp, frame: interp.Frame[types.TypeAttribute], stmt: Unpack):
        value = frame.get(stmt.value)
        if isinstance(value, types.Generic) and value.is_subseteq(types.Tuple):
            if value.vararg:
                rest = tuple(value.vararg.typ for _ in stmt.names[len(value.vars) :])
                return tuple(value.vars) + rest
            else:
                return value.vars
        # TODO: support unpacking other types
        return tuple(types.Any for _ in stmt.names)


def unpacking(state: lowering.LoweringState, node: ast.expr, value: ir.SSAValue):
    if isinstance(node, ast.Name):
        state.current_frame.defs[node.id] = value
        value.name = node.id
        return
    elif not isinstance(node, ast.Tuple):
        raise DialectLoweringError(f"unsupported unpack node {node}")

    names: list[str | None] = []
    continue_unpack: list[int] = []
    for idx, item in enumerate(node.elts):
        if isinstance(item, ast.Name):
            names.append(item.id)
        else:
            names.append(None)
            continue_unpack.append(idx)
    stmt = state.append_stmt(Unpack(value, tuple(names)))
    for name, result in zip(names, stmt.results):
        if name is not None:
            state.current_frame.defs[name] = result

    for idx in continue_unpack:
        unpacking(state, node.elts[idx], stmt.results[idx])



"math dialect, modeling functions in python's `math` stdlib"  # This file is generated by gen.py

from kirin.dialects.math.stmts import (
    cos as cos,
    erf as erf,
    exp as exp,
    pow as pow,
    sin as sin,
    tan as tan,
    ulp as ulp,
    acos as acos,
    asin as asin,
    atan as atan,
    ceil as ceil,
    cosh as cosh,
    erfc as erfc,
    fabs as fabs,
    fmod as fmod,
    log2 as log2,
    sinh as sinh,
    sqrt as sqrt,
    tanh as tanh,
    asinh as asinh,
    atan2 as atan2,
    atanh as atanh,
    expm1 as expm1,
    floor as floor,
    gamma as gamma,
    isinf as isinf,
    isnan as isnan,
    log1p as log1p,
    log10 as log10,
    trunc as trunc,
    lgamma as lgamma,
    degrees as degrees,
    radians as radians,
    copysign as copysign,
    isfinite as isfinite,
    remainder as remainder,
)
from kirin.dialects.math.interp import MathMethodTable as MathMethodTable
from kirin.dialects.math.dialect import dialect as dialect



"""Base dialect for Python.

This dialect does not contain statements. It only contains
lowering rules for `ast.Name` and `ast.Expr`.
"""

import ast

from kirin import ir, lowering, exceptions

dialect = ir.Dialect("py.base")


@dialect.register
class PythonLowering(lowering.FromPythonAST):

    def lower_Name(
        self, state: lowering.LoweringState, node: ast.Name
    ) -> lowering.Result:
        name = node.id
        if isinstance(node.ctx, ast.Load):
            value = state.current_frame.get(name)
            if value is None:
                raise exceptions.DialectLoweringError(f"{name} is not defined")
            return lowering.Result(value)
        elif isinstance(node.ctx, ast.Store):
            raise exceptions.DialectLoweringError("unhandled store operation")
        else:  # Del
            raise exceptions.DialectLoweringError("unhandled del operation")

    def lower_Expr(
        self, state: lowering.LoweringState, node: ast.Expr
    ) -> lowering.Result:
        return state.visit(node.value)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from kirin.lowering import LoweringState
    from kirin.ir.nodes.base import IRNode


class InterpreterExit(Exception):
    pass


class DialectDefError(Exception):
    pass


class DialectSyntaxError(Exception):
    pass


class DialectLoweringError(Exception):

    def append_hint(self, lowering: LoweringState):
        msg = self.args[0]
        source = lowering.source
        if lowering.lines:
            lines = lowering.lines
            begin = max(0, source.lineno - lowering.max_lines)
            end = min(
                len(lines),
                source.end_lineno or source.lineno + lowering.max_lines,
            )
            lines = (
                lines[begin : source.lineno]
                + [("-" * source.col_offset) + "^"]
                + lines[lowering.lineno_offset : end]
            )
            trace = "\n".join(lines)
            msg = f"{msg}: \n\n{trace}"
        else:
            msg = f"{msg}: {source.lineno}:{source.col_offset}"

        self.args = (msg,)
        return self


class CompilerError(Exception):
    pass


class VerificationError(Exception):
    def __init__(self, node: "IRNode", *messages: str) -> None:
        super().__init__(*messages)
        self.node = node


class DuplicatedDefinitionError(Exception):
    pass



from dataclasses import dataclass

from kirin import ir
from kirin.dialects import func
from kirin.rewrite.abc import RewriteRule, RewriteResult


@dataclass
class InlineGetField(RewriteRule):

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if not isinstance(node, func.GetField):
            return RewriteResult()

        if not isinstance(node.obj.owner, func.Lambda):
            return RewriteResult()

        original = node.obj.owner.captured[node.field]
        node.result.replace_by(original)
        node.delete()
        return RewriteResult(has_done_something=True)



from kirin.lowering.core import Lowering as Lowering
from kirin.lowering.frame import Frame as Frame
from kirin.lowering.state import LoweringState as LoweringState
from kirin.lowering.result import Result as Result
from kirin.lowering.stream import StmtStream as StmtStream
from kirin.lowering.binding import wraps as wraps
from kirin.lowering.dialect import FromPythonAST as FromPythonAST
